import dataclasses
import datetime
from collections.abc import Callable

import dacite
import numpy as np
import pytest
import torch

from fme.core import AtmosphereData, metrics
from fme.core.coordinates import HybridSigmaPressureCoordinate
from fme.core.derived_variables import total_water_path_budget_residual
from fme.core.gridded_ops import GriddedOperations, HEALPixOperations, LatLonOperations
from fme.core.typing_ import TensorMapping

from .atmosphere import (
    AtmosphereCorrector,
    AtmosphereCorrectorConfig,
    EnergyBudgetConfig,
    _force_conserve_dry_air,
    _force_conserve_moisture,
    _force_conserve_total_energy,
    _force_zero_global_mean_moisture_advection,
)
from .utils import force_positive

TIMESTEP = datetime.timedelta(hours=6)


def test_config_dataclass_round_trip():
    config = AtmosphereCorrectorConfig()
    new_config = dacite.from_dict(
        data_class=AtmosphereCorrectorConfig,
        data=dataclasses.asdict(config),
        config=dacite.Config(strict=True),
    )
    assert config == new_config


def compute_dry_air_absolute_differences(
    atmosphere_data: AtmosphereData,
    area_weighted_mean: Callable[[torch.Tensor], torch.Tensor],
) -> torch.Tensor:
    """
    Computes the absolute value of the dry air tendency of each time step.

    Args:
        atmosphere_data: AtmosphereData object.
        area_weighted_mean: Function which returns an area-weighted mean.
        vertical_coordinate: The vertical coordinate of the model.

    Returns:
        A tensor of shape (time,) of the absolute value of the dry air tendency
            of each time step.
    """
    try:
        surface_pressure = atmosphere_data.surface_pressure
        total_water_path = atmosphere_data.total_water_path
    except KeyError:
        return torch.tensor([torch.nan])
    ps_dry = metrics.surface_pressure_due_to_dry_air(surface_pressure, total_water_path)
    ps_dry_mean = area_weighted_mean(ps_dry)
    return ps_dry_mean.diff(dim=-1).abs().mean(dim=0)


def get_dry_air_nonconservation(
    data: TensorMapping,
    area_weighted_mean: Callable[[torch.Tensor], torch.Tensor],
    vertical_coordinate: HybridSigmaPressureCoordinate,
):
    """
    Computes the time-average one-step absolute difference in surface pressure due to
    changes in globally integrated dry air.

    Args:
        data: A mapping from variable name to tensor of shape
            [sample, time, lat, lon], in physical units. specific_total_water in kg/kg
            and surface_pressure in Pa must be present.
        area_weighted_mean: Computes the area-weighted mean of a tensor, removing the
            horizontal dimensions.
        vertical_coordinate: The vertical coordinates of the model.
    """
    return compute_dry_air_absolute_differences(
        AtmosphereData(data, vertical_coordinate),
        area_weighted_mean=area_weighted_mean,
    ).mean()


def test_force_no_global_mean_moisture_advection():
    torch.random.manual_seed(0)
    data = {
        "tendency_of_total_water_path_due_to_advection": torch.rand(size=(3, 2, 5, 5)),
    }
    area_weights = 1.0 + torch.rand(size=(5, 1)).broadcast_to(size=(5, 5))
    original_mean = metrics.weighted_mean(
        data["tendency_of_total_water_path_due_to_advection"],
        weights=area_weights,
        dim=[-2, -1],
    )
    assert (original_mean.abs() > 0.0).all()
    fixed_data = _force_zero_global_mean_moisture_advection(
        data,
        area_weighted_mean=LatLonOperations(area_weights).area_weighted_mean,
    )
    new_mean = metrics.weighted_mean(
        fixed_data["tendency_of_total_water_path_due_to_advection"],
        weights=area_weights,
        dim=[-2, -1],
    )
    assert (new_mean.abs() < original_mean.abs()).all()
    np.testing.assert_almost_equal(new_mean.cpu().numpy(), 0.0, decimal=6)


@pytest.mark.parametrize(
    "size, use_area",
    [
        pytest.param((3, 2, 5, 5), True, id="latlon"),
        pytest.param((3, 12, 2, 3, 3), False, id="healpix"),
    ],
)
def test_force_conserve_dry_air(size: tuple[int, ...], use_area: bool):
    torch.random.manual_seed(0)
    data = {
        "PRESsfc": 10.0 + torch.rand(size=size),
        "specific_total_water_0": torch.rand(size=size),
        "specific_total_water_1": torch.rand(size=size),
    }
    vertical_coordinate = HybridSigmaPressureCoordinate(
        ak=torch.asarray([3.0, 1.0, 0.0]), bk=torch.asarray([0.0, 0.6, 1.0])
    )
    if use_area:
        area_weights: torch.Tensor | None = 1.0 + torch.rand(
            size=(size[-2], 1)
        ).broadcast_to(size=size[-2:])
    else:
        area_weights = None
    if area_weights is not None:
        gridded_operations: GriddedOperations = LatLonOperations(area_weights)
    else:
        gridded_operations = HEALPixOperations()
    original_nonconservation = get_dry_air_nonconservation(
        data,
        vertical_coordinate=vertical_coordinate,
        area_weighted_mean=gridded_operations.area_weighted_mean,
    )
    assert original_nonconservation > 0.0
    in_data = {k: v.select(dim=1, index=0) for k, v in data.items()}
    out_data = {k: v.select(dim=1, index=1) for k, v in data.items()}
    fixed_out_data = _force_conserve_dry_air(
        in_data,
        out_data,
        vertical_coordinate=vertical_coordinate,
        area_weighted_mean=gridded_operations.area_weighted_mean,
    )
    new_data = {
        k: torch.stack([v, fixed_out_data[k]], dim=1) for k, v in in_data.items()
    }
    new_nonconservation = get_dry_air_nonconservation(
        new_data,
        vertical_coordinate=vertical_coordinate,
        area_weighted_mean=gridded_operations.area_weighted_mean,
    )
    assert new_nonconservation < original_nonconservation
    np.testing.assert_almost_equal(new_nonconservation.cpu().numpy(), 0.0, decimal=6)


@pytest.mark.parametrize("dataset", ["fv3", "e3sm"])
@pytest.mark.parametrize(
    "global_only, terms_to_modify",
    [
        (True, "precipitation"),
        (True, "evaporation"),
        (False, "advection_and_precipitation"),
        (False, "advection_and_evaporation"),
    ],
)
@pytest.mark.parametrize(
    "size, use_area",
    [
        pytest.param((3, 2, 5, 5), True, id="latlon"),
        pytest.param((3, 12, 2, 3, 3), False, id="healpix"),
    ],
)
def test_force_conserve_moisture(
    dataset: str,
    global_only: bool,
    terms_to_modify,
    size: tuple[int, ...],
    use_area: bool,
):
    torch.random.manual_seed(0)
    if dataset == "fv3":
        data = {
            "PRESsfc": 10.0 + torch.rand(size=size),
            "specific_total_water_0": torch.rand(size=size),
            "specific_total_water_1": torch.rand(size=size),
            "PRATEsfc": torch.rand(size=size),
            "LHTFLsfc": torch.rand(size=size),
            "tendency_of_total_water_path_due_to_advection": torch.rand(size=size),
        }
    if dataset == "e3sm":
        data = {
            "PS": 10.0 + torch.rand(size=size),
            "specific_total_water_0": torch.rand(size=size),
            "specific_total_water_1": torch.rand(size=size),
            "surface_precipitation_rate": torch.rand(size=size),
            "LHFLX": torch.rand(size=size),
            "tendency_of_total_water_path_due_to_advection": torch.rand(size=size),
        }
    vertical_coordinate = HybridSigmaPressureCoordinate(
        ak=torch.asarray([3.0, 1.0, 0.0]), bk=torch.asarray([0.0, 0.6, 1.0])
    )
    if use_area:
        ops: GriddedOperations = LatLonOperations(
            1.0 + torch.rand(size=(5, 1)).broadcast_to(size=(5, 5))
        )
    else:
        ops = HEALPixOperations()
    data["tendency_of_total_water_path_due_to_advection"] -= ops.area_weighted_mean(
        data["tendency_of_total_water_path_due_to_advection"], keepdim=True
    )
    original_budget_residual = total_water_path_budget_residual(
        AtmosphereData(data, vertical_coordinate),
        timestep=TIMESTEP,
    )[:, 1]  # no meaning for initial value data, want first timestep
    if global_only:
        original_budget_residual = ops.area_weighted_mean(
            original_budget_residual, keepdim=True
        )
    original_budget_residual = original_budget_residual.cpu().numpy()
    original_dry_air = (
        AtmosphereData(data, vertical_coordinate)
        .surface_pressure_due_to_dry_air.cpu()
        .numpy()
    )
    assert np.any(np.abs(original_budget_residual) > 0.0)
    in_data = {k: v.select(dim=1, index=0) for k, v in data.items()}
    out_data = {k: v.select(dim=1, index=1) for k, v in data.items()}
    fixed_out_data = _force_conserve_moisture(
        in_data,
        out_data,
        vertical_coordinate=vertical_coordinate,
        area_weighted_mean=ops.area_weighted_mean,
        timestep_seconds=TIMESTEP.total_seconds(),
        terms_to_modify=terms_to_modify,
    )
    new_data = {
        k: torch.stack([v, fixed_out_data[k]], dim=1) for k, v in in_data.items()
    }
    new_budget_residual = total_water_path_budget_residual(
        AtmosphereData(new_data, vertical_coordinate),
        timestep=TIMESTEP,
    )[:, 1]  # no meaning for initial value data, want first timestep
    new_dry_air = (
        AtmosphereData(data, vertical_coordinate)
        .surface_pressure_due_to_dry_air.cpu()
        .numpy()
    )

    global_budget_residual = ops.area_weighted_mean(new_budget_residual).cpu().numpy()
    np.testing.assert_almost_equal(global_budget_residual, 0.0, decimal=6)

    if not global_only:
        new_budget_residual = new_budget_residual.cpu().numpy()
        assert np.all(np.abs(new_budget_residual) < np.abs(original_budget_residual))
        np.testing.assert_almost_equal(new_budget_residual, 0.0, decimal=6)

    np.testing.assert_almost_equal(new_dry_air, original_dry_air, decimal=6)


def test_force_positive():
    data = {
        "foo": torch.tensor([[-1.0, 0.0], [0.0, 1.0], [1.0, 2.0]]),
        "bar": torch.tensor([[-1.0, 0.0], [0.0, -3.0], [1.0, 2.0]]),
    }
    original_min = torch.min(data["foo"])
    assert original_min < 0.0
    fixed_data = force_positive(data, ["foo"])
    new_min = torch.min(fixed_data["foo"])
    # Ensure the minimum value of 'foo' is now 0
    torch.testing.assert_close(new_min, torch.tensor(0.0))
    # Ensure other variables are not modified
    torch.testing.assert_close(fixed_data["bar"], data["bar"])


def _get_corrector_test_input(tensor_shape):
    """Generate test input that has necessary variables for all correctors."""
    # debugging a bit easier with realistic scales for input variables
    # tuples of (name, mean, std) for each variable
    variables = [
        ("PRESsfc", 100000, 1000),
        ("HGTsfc", 500, 50),
        ("SHTFLsfc", 100, 50),
        ("LHTFLsfc", 100, 50),
        ("specific_total_water_0", 0.001, 0.0001),
        ("specific_total_water_1", 0.001, 0.0001),
        ("air_temperature_0", 300, 10),
        ("air_temperature_1", 300, 10),
        ("DSWRFtoa", 500, 100),
        ("USWRFtoa", 500, 100),
        ("ULWRFtoa", 500, 100),
        ("USWRFsfc", 500, 100),
        ("DSWRFsfc", 500, 100),
        ("ULWRFsfc", 500, 100),
        ("DLWRFsfc", 500, 100),
        ("PRATEsfc", 1e-5, 1e-6),
        ("tendency_of_total_water_path_due_to_advection", 1e-5, 1e-6),
    ]
    forcing_names = ["HGTsfc", "DSWRFtoa"]
    non_forcing_names = [name for name, _, _ in variables if name not in forcing_names]
    input_data = {
        name: centering + scale * torch.rand(size=tensor_shape)
        for name, centering, scale in variables
    }
    gen_data = {
        name: centering + scale * torch.rand(size=tensor_shape)
        for name, centering, scale in variables
        if name in non_forcing_names
    }
    forcing_data = {
        name: centering + scale * torch.rand(size=tensor_shape)
        for name, centering, scale in variables
        if name in forcing_names
    }
    vertical_coord = HybridSigmaPressureCoordinate(
        ak=torch.asarray([3.0, 1.0, 0.0]), bk=torch.asarray([0.0, 0.6, 1.0])
    )
    return input_data, gen_data, forcing_data, vertical_coord


@pytest.mark.parametrize("negative_pressure", [True, False])
def test__force_conserve_total_energy(negative_pressure: bool):
    tensor_shape = (5, 5)

    ops = LatLonOperations(
        0.5 + torch.rand(size=(tensor_shape[-2], 1)).broadcast_to(size=tensor_shape)
    )
    timestep = datetime.timedelta(seconds=3600)
    input_data, gen_data, forcing_data, vertical_coord = _get_corrector_test_input(
        tensor_shape
    )
    extra_heating = 10.0

    if negative_pressure:
        input_data["PRESsfc"] = -1 * input_data["PRESsfc"]

    correction_expected = not negative_pressure

    corrected_gen_data = _force_conserve_total_energy(
        input_data=input_data,
        gen_data=gen_data,
        forcing_data=forcing_data,
        area_weighted_mean=ops.area_weighted_mean,
        vertical_coordinate=vertical_coord,
        timestep_seconds=timestep.total_seconds(),
        unaccounted_heating=extra_heating,
    )

    # ensure only temperature is modified
    for name in gen_data:
        if "air_temperature" in name:
            if correction_expected:
                assert not torch.allclose(
                    corrected_gen_data[name], gen_data[name], rtol=1e-6
                )
            else:
                assert torch.allclose(
                    corrected_gen_data[name], gen_data[name], rtol=1e-6
                )
        else:
            torch.testing.assert_close(corrected_gen_data[name], gen_data[name])

    # ensure forcing variables are not in the corrected data
    for name in forcing_data:
        assert name not in corrected_gen_data

    # ensure the corrected global mean MSE path is what we expect
    input = AtmosphereData(input_data, vertical_coord)
    corrected_gen = AtmosphereData(corrected_gen_data | forcing_data, vertical_coord)
    input_gm_mse = ops.area_weighted_mean(input.total_energy_ace2_path)
    corrected_gen_gm_mse = ops.area_weighted_mean(corrected_gen.total_energy_ace2_path)
    predicted_mse_tendency = (
        ops.area_weighted_mean(corrected_gen.net_energy_flux_into_atmosphere)
        + extra_heating
    )
    if correction_expected:
        expected_gm_mse = (
            input_gm_mse + predicted_mse_tendency * timestep.total_seconds()
        )
        torch.testing.assert_close(corrected_gen_gm_mse, expected_gm_mse)

    # ensure the temperature correction is constant
    corrected_gen_temperature = corrected_gen_data["air_temperature_1"]
    initial_gen_temperature = gen_data["air_temperature_1"]
    temperature_1_correction = corrected_gen_temperature - initial_gen_temperature
    assert torch.all(torch.eq(temperature_1_correction, temperature_1_correction[0, 0]))
    temperature_correction_0 = (
        corrected_gen_data["air_temperature_0"] - gen_data["air_temperature_0"]
    )
    torch.testing.assert_close(temperature_correction_0, temperature_1_correction)


def test__force_conserve_energy_doesnt_clobber():
    tensor_shape = (5, 5)

    ops = LatLonOperations(
        0.5 + torch.rand(size=(tensor_shape[-2], 1)).broadcast_to(size=tensor_shape)
    )
    timestep = datetime.timedelta(seconds=3600)
    input_data, gen_data, forcing_data, vertical_coord = _get_corrector_test_input(
        tensor_shape
    )
    # add a prognostic variable to the forcing data
    forcing_data["PRESsfc"] = 10.0 + torch.rand(size=tensor_shape)

    corrected_gen_data = _force_conserve_total_energy(
        input_data=input_data,
        gen_data=gen_data,
        forcing_data=forcing_data,
        area_weighted_mean=ops.area_weighted_mean,
        vertical_coordinate=vertical_coord,
        timestep_seconds=timestep.total_seconds(),
    )
    torch.testing.assert_close(corrected_gen_data["PRESsfc"], gen_data["PRESsfc"])


def test_corrector_integration():
    """Ensures that the corrector can be called with all methods active
    but doesn't check results."""
    config = AtmosphereCorrectorConfig(
        conserve_dry_air=True,
        zero_global_mean_moisture_advection=True,
        moisture_budget_correction="advection_and_precipitation",
        force_positive_names=["PRESsfc"],
        total_energy_budget_correction=EnergyBudgetConfig("constant_temperature", 1.0),
    )
    tensor_shape = (5, 5)
    test_input = _get_corrector_test_input(tensor_shape)
    input_data, gen_data, forcing_data, vertical_coord = test_input
    ops = LatLonOperations(
        0.5 + torch.rand(size=(tensor_shape[-2], 1)).broadcast_to(size=tensor_shape)
    )
    timestep = datetime.timedelta(seconds=3600)
    corrector = AtmosphereCorrector(config, ops, vertical_coord, timestep)
    corrector(input_data, gen_data, forcing_data)



import contextlib
import dataclasses
import itertools
import os
import unittest.mock
from typing import Any, TypeVar, cast

import numpy as np
import pytest
import torch

from fme.ace.data_loading.gridded_data import DataLoader
from fme.core.ema import EMATracker
from fme.core.generics.aggregator import (
    AggregatorABC,
    InferenceAggregatorABC,
    InferenceLog,
    InferenceLogs,
)
from fme.core.generics.data import GriddedDataABC, InferenceDataABC
from fme.core.generics.optimization import OptimizationABC
from fme.core.generics.trainer import (
    AggregatorBuilderABC,
    CheckpointPaths,
    TrainConfigProtocol,
    Trainer,
    TrainOutputABC,
    TrainStepperABC,
)
from fme.core.logging_utils import LoggingConfig
from fme.core.optimization import Optimization
from fme.core.scheduler import SchedulerConfig
from fme.core.testing.wandb import mock_wandb
from fme.core.typing_ import Slice, TensorDict, TensorMapping


class PSType:
    pass


class BDType:
    pass


class FDType:
    pass


class SDType:
    pass


class TrainOutput(TrainOutputABC):
    def get_metrics(self) -> dict[str, torch.Tensor]:
        return {}


class TrainData(GriddedDataABC[BDType]):
    @property
    def batch_size(self) -> int:
        return 1

    @property
    def loader(self) -> DataLoader[BDType]:
        return [BDType() for _ in range(self.n_batches)]

    @property
    def n_samples(self) -> int:
        return 3

    @property
    def n_batches(self) -> int:
        return 5

    @property
    def n_forward_steps(self) -> int:
        return 1

    def __init__(self):
        self._set_epoch = unittest.mock.MagicMock()
        self._log_info = unittest.mock.MagicMock()

    def set_epoch(self, epoch: int) -> None:
        self._set_epoch(epoch)

    def log_info(self, name: str) -> None:
        self._log_info(name)

    @property
    def set_epoch_mock(self) -> unittest.mock.Mock:
        return self._set_epoch

    @property
    def log_info_mock(self) -> unittest.mock.Mock:
        return self._log_info


class InferenceData(InferenceDataABC[PSType, FDType]):
    def __init__(self, n_time_windows: int = 1):
        self.n_time_windows = n_time_windows

    @property
    def initial_condition(self) -> PSType:
        return PSType()

    @property
    def loader(self) -> DataLoader[FDType]:
        return [FDType() for _ in range(self.n_time_windows)]

    @property
    def n_window_forward_steps(self) -> int:
        return 1


class TrainStepper(TrainStepperABC[PSType, BDType, FDType, SDType, TrainOutput]):
    SelfType = TypeVar("SelfType", bound="TrainStepper")

    def __init__(
        self,
        state: dict[str, Any] | None = None,
    ):
        self._modules = torch.nn.ModuleList([torch.nn.Linear(1, 1, bias=False)])
        self._modules[0].weight.data.fill_(0.0)
        if state is not None:
            self._state = state
        else:
            self._state = {}
        self.loaded_state: dict[str, Any] | None = None

    def get_state(self) -> dict[str, Any]:
        return {**self._state, "modules": self._modules.state_dict()}

    def load_state(self, state: dict[str, Any]) -> None:
        self._state = state
        self.loaded_state = state
        self._modules.load_state_dict(state["modules"])

    @classmethod
    def from_state(cls: type[SelfType], state: dict[str, Any]) -> SelfType:
        ret = cls()
        ret.load_state(state)
        return ret

    @property
    def modules(self) -> torch.nn.ModuleList:
        return self._modules

    @property
    def n_ic_timesteps(self) -> int:
        return 1

    def normalize(self, data: TensorMapping) -> TensorDict:
        return dict(data)

    def predict_paired(
        self,
        initial_condition: PSType,
        forcing: FDType,
        compute_derived_variables: bool = False,
    ) -> tuple[SDType, PSType]:
        return SDType(), PSType()

    def train_on_batch(
        self,
        batch: BDType,
        optimization: OptimizationABC,
        compute_derived_variables: bool = False,
    ) -> TrainOutput:
        optimization.accumulate_loss(torch.tensor(float("inf")))
        optimization.step_weights()
        return TrainOutput()

    def set_train(self) -> None:
        pass

    def set_eval(self) -> None:
        pass

    def update_training_history(self, *args: Any, **kwargs: Any) -> None:
        pass


@dataclasses.dataclass
class Config:
    experiment_dir: str = "test_experiment_dir"
    checkpoint_dir: str = "test_checkpoint_dir"
    output_dir: str = "test_output_dir"
    max_epochs: int = 2
    save_checkpoint: bool = True
    validate_using_ema: bool = True
    log_train_every_n_batches: int = 1
    inference_n_forward_steps: int = 1
    checkpoint_save_epochs: Slice | None = None
    ema_checkpoint_save_epochs: Slice | None = None
    segment_epochs: int | None = None
    evaluate_before_training: bool = False

    def __post_init__(self):
        start_epoch = 0 if self.evaluate_before_training else 1
        self.get_inference_epochs = unittest.mock.MagicMock(
            return_value=[i for i in range(start_epoch, self.max_epochs + 1)]
        )


_: TrainConfigProtocol = Config()


class TrainAggregator(AggregatorABC[TrainOutput]):
    def __init__(self, train_loss: float):
        self.train_loss = train_loss

    def record_batch(self, batch: TrainOutput) -> None:
        pass

    def get_logs(self, label: str) -> dict[str, Any]:
        return {f"{label}/mean/loss": self.train_loss}

    def flush_diagnostics(self, subdir: str | None) -> None:
        pass


class ValidationAggregator(AggregatorABC[TrainOutput]):
    def __init__(self, validation_loss: float):
        self.validation_loss = validation_loss

    def record_batch(self, batch: TrainOutput) -> None:
        pass

    def get_logs(self, label: str) -> dict[str, Any]:
        return {f"{label}/mean/loss": self.validation_loss}

    def flush_diagnostics(self, subdir: str | None) -> None:
        pass


class InferenceAggregator(InferenceAggregatorABC[PSType, SDType]):
    def __init__(self, inference_loss: float):
        self.inference_loss = inference_loss

    def record_batch(self, data: SDType) -> InferenceLogs:
        return [{}]

    def record_initial_condition(self, initial_condition: PSType) -> InferenceLogs:
        return [{}]

    def get_summary_logs(self) -> InferenceLog:
        return {"time_mean_norm/rmse/channel_mean": self.inference_loss}

    def flush_diagnostics(self, subdir: str | None) -> None:
        pass


class AggregatorBuilder(AggregatorBuilderABC[PSType, TrainOutput, SDType]):
    def __init__(
        self,
        train_losses: np.ndarray,
        validation_losses: np.ndarray,
        inference_losses: np.ndarray,
    ):
        self.train_losses = train_losses
        self.validation_losses = validation_losses
        self.inference_losses = inference_losses
        self._train_calls = 0
        self._validation_calls = 0
        self._inference_calls = 0

    def get_train_aggregator(self) -> AggregatorABC[TrainOutput]:
        ret = TrainAggregator(self.train_losses[self._train_calls])
        self._train_calls += 1
        return ret

    def get_validation_aggregator(self) -> AggregatorABC[TrainOutput]:
        ret = ValidationAggregator(self.validation_losses[self._validation_calls])
        self._validation_calls += 1
        return ret

    def get_inference_aggregator(self) -> InferenceAggregatorABC[PSType, SDType]:
        ret = InferenceAggregator(self.inference_losses[self._inference_calls])
        self._inference_calls += 1
        return ret


def get_trainer(
    tmp_path: str,
    checkpoint_save_epochs: Slice | None = None,
    segment_epochs: int | None = None,
    max_epochs: int = 8,
    checkpoint_dir: str | None = None,
    stepper_state: dict[str, Any] | None = None,
    train_losses: np.ndarray | None = None,
    validation_losses: np.ndarray | None = None,
    inference_losses: np.ndarray | None = None,
    stepper_module_values: np.ndarray | None = None,
    ema_decay: float = 0.9999,
    validate_using_ema: bool = True,
    evaluate_before_training: bool = False,
) -> tuple[TrainConfigProtocol, Trainer]:
    if checkpoint_dir is None:
        checkpoint_dir = os.path.join(tmp_path, "checkpoints")
    if train_losses is None:
        train_losses = np.zeros(max_epochs)
    if validation_losses is None:
        n_validation_steps = max_epochs + 1 if evaluate_before_training else max_epochs
        validation_losses = np.zeros(n_validation_steps)
    if inference_losses is None:
        n_inference_steps = max_epochs + 1 if evaluate_before_training else max_epochs
        inference_losses = np.zeros(n_inference_steps)
    if stepper_module_values is None:
        stepper_module_values = np.zeros(max_epochs)
    train_data = TrainData()
    validation_data = TrainData()
    inference_data = InferenceData()
    stepper = TrainStepper(state=stepper_state)

    def build_optimization(modules: torch.nn.ModuleList) -> Optimization:
        if len(modules) != 1:
            raise ValueError("Expected 1 linear module with 1 weight")
        if not isinstance(modules[0], torch.nn.Linear):
            raise ValueError("Expected a linear module")
        module = modules[0]
        if module.weight.numel() != 1:
            raise ValueError("Expected a linear module with 1 weight")
        i = 0

        opt = Optimization(
            parameters=itertools.chain(*[module.parameters() for module in modules]),
            optimizer_type="Adam",
            lr=0.01,
            max_epochs=max_epochs,
            scheduler=SchedulerConfig(),
            enable_automatic_mixed_precision=False,
            kwargs={},
        )
        original_step_scheduler = opt.step_scheduler

        def step_scheduler_side_effect(*args, **kwargs):
            original_step_scheduler(*args, **kwargs)
            nonlocal i
            i += 1

        opt.step_scheduler = unittest.mock.MagicMock(  # type: ignore
            side_effect=step_scheduler_side_effect
        )

        def step_weights_side_effect(*args, **kwargs):
            if stepper_module_values is None:
                raise ValueError("stepper_module_values is None")
            module.weight.data.fill_(stepper_module_values[i])

        opt.step_weights = unittest.mock.MagicMock(side_effect=step_weights_side_effect)  # type: ignore
        return opt

    def build_ema(modules: torch.nn.ModuleList) -> EMATracker:
        return EMATracker(modules, decay=ema_decay)

    config: TrainConfigProtocol = Config(
        experiment_dir=tmp_path,
        checkpoint_dir=checkpoint_dir,
        checkpoint_save_epochs=checkpoint_save_epochs,
        segment_epochs=segment_epochs,
        max_epochs=max_epochs,
        validate_using_ema=validate_using_ema,
        evaluate_before_training=evaluate_before_training,
    )
    aggregator_builder = AggregatorBuilder(
        train_losses=train_losses,
        validation_losses=validation_losses,
        inference_losses=inference_losses,
    )
    return config, Trainer(
        train_data=train_data,
        validation_data=validation_data,
        inference_data=inference_data,
        stepper=stepper,
        build_optimization=build_optimization,
        build_ema=build_ema,
        config=config,
        aggregator_builder=aggregator_builder,
        end_of_batch_callback=unittest.mock.MagicMock(),
        end_of_epoch_callback=unittest.mock.MagicMock(side_effect=lambda epoch: {}),
        do_gc_collect=False,  # for much faster tests
    )


@pytest.mark.parametrize(
    "checkpoint_save_epochs",
    [None, Slice(start=2, stop=3), Slice(start=1, step=2)],
)
def test_trainer(tmp_path: str, checkpoint_save_epochs: Slice | None):
    config, trainer = get_trainer(tmp_path, checkpoint_save_epochs, max_epochs=4)
    trainer.train()
    assert os.path.exists(config.experiment_dir)
    assert os.path.exists(config.checkpoint_dir)
    paths = CheckpointPaths(config.checkpoint_dir)
    assert os.path.exists(paths.latest_checkpoint_path)
    assert os.path.exists(paths.best_checkpoint_path)
    assert os.path.exists(paths.best_inference_checkpoint_path)
    assert os.path.exists(paths.ema_checkpoint_path)
    save_epochs = list(range(config.max_epochs))
    if checkpoint_save_epochs is not None:
        save_epochs = save_epochs[checkpoint_save_epochs.slice]
    else:
        save_epochs = []
    for i in range(config.max_epochs):
        if i in save_epochs:
            assert os.path.exists(paths.epoch_checkpoint_path(i))
        else:
            assert not os.path.exists(paths.epoch_checkpoint_path(i))
        assert not os.path.exists(paths.ema_epoch_checkpoint_path(i))
    train_data = cast(TrainData, trainer.train_data)
    valid_data = cast(TrainData, trainer.valid_data)
    assert train_data.set_epoch_mock.mock_calls == [
        unittest.mock.call(i) for i in range(1, config.max_epochs + 1)
    ]
    assert valid_data.set_epoch_mock.mock_calls == []  # no shuffling
    assert train_data.log_info_mock.called
    assert valid_data.log_info_mock.called
    assert trainer._end_of_epoch_callback.mock_calls == [  # type: ignore
        unittest.mock.call(i) for i in range(1, config.max_epochs + 1)
    ]


@pytest.mark.parametrize("segment_epochs", [1, 2, 3])
def test_segmented_trainer_runs_correct_epochs(tmp_path: str, segment_epochs: int):
    max_epochs = 4
    total_segments = max_epochs // segment_epochs
    for i in range(total_segments):
        config, trainer = get_trainer(
            tmp_path,
            checkpoint_dir=os.path.join(tmp_path, "checkpoint_dir"),  # same dir for all
            # for speed, don't save per-epoch checkpoints for this test
            checkpoint_save_epochs=Slice(start=0, stop=0),
            segment_epochs=segment_epochs,
            max_epochs=max_epochs,
        )
        trainer.train()
        paths = CheckpointPaths(config.checkpoint_dir)
        assert os.path.exists(paths.latest_checkpoint_path)
        train_data = cast(TrainData, trainer.train_data)
        assert train_data.set_epoch_mock.mock_calls == [
            unittest.mock.call(i)
            for i in range(
                i * segment_epochs + 1,
                min((i + 1) * segment_epochs, config.max_epochs) + 1,
            )
        ]


class TrainingInterrupted(Exception):
    pass


@contextlib.contextmanager
def fail_after_calls_patch(object, method: str, call_count: int):
    total_calls = 0
    original_method = getattr(object, method)

    def wrapper(*args, **kwargs):
        nonlocal total_calls
        total_calls += 1
        if total_calls >= call_count:
            raise TrainingInterrupted()
        return original_method(*args, **kwargs)

    with unittest.mock.patch.object(object, method) as mock:
        mock.side_effect = wrapper
        try:
            yield mock
        except TrainingInterrupted:
            pass


@pytest.mark.parametrize(
    "interrupt_method",
    ["train_one_epoch", "validate_one_epoch", "inference_one_epoch"],
)
def test_resume_after_interrupted_training(tmp_path: str, interrupt_method: str):
    max_epochs = 4
    calls_before_interrupt = 2
    stepper_state = {"foo": "bar"}
    config, trainer = get_trainer(
        tmp_path,
        stepper_state=stepper_state,
        checkpoint_save_epochs=Slice(start=0, stop=0),
        max_epochs=max_epochs,
    )
    with fail_after_calls_patch(trainer, interrupt_method, calls_before_interrupt):
        trainer.train()
    train_data = cast(TrainData, trainer.train_data)
    assert train_data.set_epoch_mock.mock_calls == [
        unittest.mock.call(i) for i in range(1, calls_before_interrupt + 1)
    ]
    paths = CheckpointPaths(config.checkpoint_dir)
    assert os.path.exists(paths.latest_checkpoint_path)
    _, trainer = get_trainer(
        tmp_path,
        checkpoint_save_epochs=Slice(start=0, stop=0),
        max_epochs=max_epochs,
        stepper_state=stepper_state,
    )
    trainer.train()
    train_data = cast(TrainData, trainer.train_data)
    assert train_data.set_epoch_mock.mock_calls == [
        unittest.mock.call(i) for i in range(calls_before_interrupt, max_epochs + 1)
    ]
    stepper = cast(TrainStepper, trainer.stepper)
    assert stepper.loaded_state is not None
    assert stepper.loaded_state["foo"] == "bar"
    assert "modules" in stepper.loaded_state
    assert len(stepper.loaded_state) == 2


@pytest.mark.parametrize("ema_decay", [0.05, 0.99])
@pytest.mark.parametrize("validate_using_ema", [True, False])
def test_saves_correct_ema_checkpoints(
    tmp_path: str, ema_decay: float, validate_using_ema: bool
):
    config, trainer = get_trainer(
        tmp_path,
        checkpoint_dir=os.path.join(tmp_path, "checkpoint_dir"),  # same dir for all
        ema_decay=ema_decay,
        validate_using_ema=validate_using_ema,
    )
    valid_loss = 0.1
    inference_error = 0.2
    trainer.stepper.modules[0].weight.data.fill_(1.0)
    trainer._ema(model=trainer.stepper.modules)
    trainer.save_all_checkpoints(valid_loss=valid_loss, inference_error=inference_error)
    paths = CheckpointPaths(config.checkpoint_dir)
    assert os.path.exists(paths.ema_checkpoint_path)
    ema_checkpoint = torch.load(paths.ema_checkpoint_path)
    ema_weight = 1.0 - min(ema_decay, 2.0 / 11.0)
    np.testing.assert_allclose(
        ema_checkpoint["stepper"]["modules"]["0.weight"].cpu().numpy(),
        ema_weight,
        atol=1e-7,
    )
    assert ema_checkpoint["best_validation_loss"] == valid_loss
    assert ema_checkpoint["best_inference_error"] == inference_error
    assert os.path.exists(paths.latest_checkpoint_path)
    latest_checkpoint = torch.load(paths.latest_checkpoint_path)
    np.testing.assert_allclose(
        latest_checkpoint["stepper"]["modules"]["0.weight"].cpu().numpy(),
        1.0,
        atol=1e-7,
    )
    assert latest_checkpoint["best_validation_loss"] == valid_loss
    assert latest_checkpoint["best_inference_error"] == inference_error
    if validate_using_ema:
        best_weight = ema_weight
    else:
        best_weight = 1.0
    assert os.path.exists(paths.best_checkpoint_path)
    best_checkpoint = torch.load(paths.best_checkpoint_path)
    assert best_checkpoint["best_validation_loss"] == valid_loss
    assert best_checkpoint["best_inference_error"] == inference_error
    np.testing.assert_allclose(
        best_checkpoint["stepper"]["modules"]["0.weight"].cpu().numpy(),
        best_weight,
        atol=1e-7,
    )
    best_inference_checkpoint = torch.load(
        paths.best_inference_checkpoint_path, weights_only=False
    )
    assert best_inference_checkpoint["best_validation_loss"] == valid_loss
    assert best_inference_checkpoint["best_inference_error"] == inference_error
    np.testing.assert_allclose(
        best_inference_checkpoint["stepper"]["modules"]["0.weight"].cpu().numpy(),
        best_weight,
        atol=1e-7,
    )


@pytest.mark.parametrize(
    "segment_epochs, best_val_epoch, best_inference_epoch",
    [
        (None, 1, 1),
        (None, 3, 5),
        (2, 3, 5),
        (2, 5, 3),
        (2, 4, 6),
        (2, 6, 4),
    ],
)
def test_saves_correct_non_ema_epoch_checkpoints(
    tmp_path: str,
    segment_epochs: int | None,
    best_val_epoch: int,
    best_inference_epoch: int,
):
    max_epochs = 10
    if segment_epochs is None:
        total_segments = 1
        segment_epochs_value = max_epochs
    else:
        total_segments = max_epochs // segment_epochs
        segment_epochs_value = segment_epochs
    train_losses = np.random.rand(max_epochs) + 0.01
    val_losses = np.random.rand(max_epochs) + 0.01
    inference_losses = np.random.rand(max_epochs) + 0.01
    val_losses[best_val_epoch - 1] = 0.0
    inference_losses[best_inference_epoch - 1] = 0.0
    module_values = np.random.rand(max_epochs)
    for i in range(total_segments):
        config, trainer = get_trainer(
            tmp_path,
            checkpoint_dir=os.path.join(tmp_path, "checkpoint_dir"),  # same dir for all
            # for speed, don't save per-epoch checkpoints for this test
            checkpoint_save_epochs=Slice(start=0, stop=0),
            segment_epochs=segment_epochs,
            max_epochs=max_epochs,
            train_losses=train_losses[
                i * segment_epochs_value : (i + 1) * segment_epochs_value
            ],
            validation_losses=val_losses[
                i * segment_epochs_value : (i + 1) * segment_epochs_value
            ],
            inference_losses=inference_losses[
                i * segment_epochs_value : (i + 1) * segment_epochs_value
            ],
            stepper_module_values=module_values[
                i * segment_epochs_value : (i + 1) * segment_epochs_value
            ],
            validate_using_ema=False,
        )
        trainer.train()
        paths = CheckpointPaths(config.checkpoint_dir)
        assert os.path.exists(paths.latest_checkpoint_path)
        train_data = cast(TrainData, trainer.train_data)
        assert train_data.set_epoch_mock.mock_calls == [
            unittest.mock.call(i)
            for i in range(
                i * segment_epochs_value + 1,
                min((i + 1) * segment_epochs_value, config.max_epochs) + 1,
            )
        ]
        latest_checkpoint = torch.load(paths.latest_checkpoint_path, weights_only=False)
        assert latest_checkpoint["epoch"] == min(
            max_epochs, (i + 1) * segment_epochs_value
        )
        np.testing.assert_allclose(
            latest_checkpoint["stepper"]["modules"]["0.weight"].cpu().numpy(),
            module_values[min((i + 1) * segment_epochs_value - 1, max_epochs - 1)],
        )
    paths = CheckpointPaths(config.checkpoint_dir)
    assert os.path.exists(paths.latest_checkpoint_path)
    assert os.path.exists(paths.best_checkpoint_path)
    assert os.path.exists(paths.best_inference_checkpoint_path)
    assert os.path.exists(paths.ema_checkpoint_path)
    best_checkpoint = torch.load(paths.best_checkpoint_path, weights_only=False)
    assert best_checkpoint["epoch"] == best_val_epoch
    assert best_checkpoint["best_validation_loss"] == 0.0
    assert best_checkpoint["best_inference_error"] == np.min(
        inference_losses[:best_val_epoch]
    )
    np.testing.assert_allclose(
        best_checkpoint["stepper"]["modules"]["0.weight"].cpu().numpy(),
        module_values[best_val_epoch - 1],
    )
    best_inference_checkpoint = torch.load(
        paths.best_inference_checkpoint_path, weights_only=False
    )
    assert best_inference_checkpoint["epoch"] == best_inference_epoch
    assert best_inference_checkpoint["best_validation_loss"] == np.min(
        val_losses[:best_inference_epoch]
    )
    assert best_inference_checkpoint["best_inference_error"] == 0.0
    latest_checkpoint = torch.load(paths.latest_checkpoint_path, weights_only=False)
    assert latest_checkpoint["epoch"] == max_epochs
    np.testing.assert_allclose(
        latest_checkpoint["stepper"]["modules"]["0.weight"].cpu().numpy(),
        module_values[-1],
    )


def test_evaluate_before_training(tmp_path: str):
    max_epochs = 2
    n_batches = TrainData().n_batches
    train_losses = np.random.rand(max_epochs)
    val_losses = np.random.rand(max_epochs + 1)
    inference_errors = np.random.rand(max_epochs + 1)
    train_loss_name = "train/mean/loss"
    val_loss_name = "val/mean/loss"
    inference_error_name = "inference/time_mean_norm/rmse/channel_mean"

    def _get_trainer(train_losses, val_losses, inference_errors):
        _, trainer = get_trainer(
            tmp_path,
            max_epochs=max_epochs,
            evaluate_before_training=True,
            train_losses=train_losses,
            validation_losses=val_losses,
            inference_losses=inference_errors,
            segment_epochs=1,
        )
        return trainer

    with mock_wandb() as wandb:
        LoggingConfig(log_to_wandb=True).configure_wandb({"experiment_dir": tmp_path})
        # run training in two segments to ensure coverage of check that extra validation
        # really only happens before any training is done.
        trainer = _get_trainer(train_losses[:1], val_losses[:2], inference_errors[:2])
        trainer.train()
        trainer = _get_trainer(train_losses[1:], val_losses[2:], inference_errors[2:])
        trainer.train()  # job is segmented, so need to call train twice to complete
        wandb_logs = wandb.get_logs()

        for i in range(max_epochs + 1):
            # only validate logs at end of each epoch, not per-batch logs
            logs = wandb_logs[i * n_batches]
            assert logs["epoch"] == i
            assert logs[val_loss_name] == val_losses[i]
            assert logs[inference_error_name] == inference_errors[i]
            if i == 0:
                assert train_loss_name not in logs
            else:
                assert logs[train_loss_name] == train_losses[i - 1]



import dataclasses
from collections.abc import Mapping

import numpy as np
import torch
import xarray as xr

from fme.core.dataset.data_typing import VariableMetadata
from fme.core.distributed import Distributed
from fme.core.typing_ import TensorDict, TensorMapping
from fme.core.wandb import WandB

wandb = WandB.get_instance()


def _get_gen_shape(gen_data: TensorMapping):
    for name in gen_data:
        return gen_data[name].shape
    raise ValueError("No data in gen_data")


@dataclasses.dataclass
class _ErrorData:
    rmse: TensorDict
    min_err: TensorDict
    max_err: TensorDict


class _ErrorVideoData:
    """
    Record batches of video data and compute statistics on the error.
    """

    def __init__(self, n_timesteps: int):
        self._mse_data: TensorDict | None = None
        self._min_err_data: TensorDict | None = None
        self._max_err_data: TensorDict | None = None
        self._n_timesteps = n_timesteps
        self._n_batches = torch.zeros([n_timesteps], dtype=torch.int32).cpu()
        self._dist = Distributed.get_instance()

    @torch.no_grad()
    def record_batch(
        self,
        target_data: TensorMapping,
        gen_data: TensorMapping,
        i_time_start: int,
    ):
        """
        Record a batch of data.

        Args:
            target_data: Dict of tensors of shape (n_samples, n_timesteps, ...)
            gen_data: Dict of tensors of shape (n_samples, n_timesteps, ...)
            i_time_start: Index of the first timestep in the batch.
        """
        if self._mse_data is None:
            self._mse_data = _initialize_video_from_batch(gen_data, self._n_timesteps)
        if self._min_err_data is None:
            self._min_err_data = _initialize_video_from_batch(
                gen_data, self._n_timesteps, fill_value=np.inf
            )
        if self._max_err_data is None:
            self._max_err_data = _initialize_video_from_batch(
                gen_data, self._n_timesteps, fill_value=-np.inf
            )

        window_steps = next(iter(target_data.values())).shape[1]
        time_slice = slice(i_time_start, i_time_start + window_steps)
        for name, gen_tensor in gen_data.items():
            target_tensor = target_data[name]
            error_tensor = (gen_tensor - target_tensor).cpu()
            self._mse_data[name][time_slice, ...] += torch.mean(
                torch.square(error_tensor), dim=0
            )
            self._min_err_data[name][time_slice, ...] = torch.minimum(
                self._min_err_data[name][time_slice, ...], error_tensor.min(dim=0)[0]
            )
            self._max_err_data[name][time_slice, ...] = torch.maximum(
                self._max_err_data[name][time_slice, ...], error_tensor.max(dim=0)[0]
            )

        self._n_batches[time_slice] += 1

    @torch.no_grad()
    def get(
        self,
    ) -> _ErrorData:
        if (
            self._mse_data is None
            or self._min_err_data is None
            or self._max_err_data is None
        ):
            raise RuntimeError("No data recorded")
        rmse_data = {}
        min_err_data = {}
        max_err_data = {}
        for name in sorted(self._mse_data):
            tensor = self._mse_data[name]
            mse = (tensor / self._n_batches[None, :, None, None]).mean(dim=0)
            mse = self._dist.reduce_mean(mse)
            rmse_data[name] = torch.sqrt(mse)
        for name in sorted(self._min_err_data):
            min_err_data[name] = self._dist.reduce_min(self._min_err_data[name])
        for name in sorted(self._max_err_data):
            max_err_data[name] = self._dist.reduce_max(self._max_err_data[name])
        return _ErrorData(rmse_data, min_err_data, max_err_data)


class _MeanVideoData:
    """
    Record batches of video data and compute the mean.
    """

    def __init__(self, n_timesteps: int):
        self._target_data: TensorDict | None = None
        self._gen_data: TensorDict | None = None
        self._n_timesteps = n_timesteps
        self._n_batches = torch.zeros([n_timesteps], dtype=torch.int32).cpu()
        self._dist = Distributed.get_instance()

    @torch.no_grad()
    def record_batch(
        self,
        target_data: TensorMapping,
        gen_data: TensorMapping,
        i_time_start: int,
    ):
        """
        Record a batch of data.

        Args:
            target_data: Dict of tensors of shape (n_samples, n_timesteps, ...)
            gen_data: Dict of tensors of shape (n_samples, n_timesteps, ...)
            i_time_start: Index of the first timestep in the batch.
        """
        if self._target_data is None:
            self._target_data = _initialize_video_from_batch(
                target_data, self._n_timesteps
            )
        if self._gen_data is None:
            self._gen_data = _initialize_video_from_batch(gen_data, self._n_timesteps)

        window_steps = next(iter(target_data.values())).shape[1]
        time_slice = slice(i_time_start, i_time_start + window_steps)
        for name, tensor in target_data.items():
            self._target_data[name][time_slice, ...] += tensor.mean(dim=0).cpu()
        for name, tensor in gen_data.items():
            self._gen_data[name][time_slice, ...] += tensor.mean(dim=0).cpu()

        self._n_batches[time_slice] += 1

    @torch.no_grad()
    def get(self) -> tuple[TensorDict, TensorDict]:
        if self._gen_data is None or self._target_data is None:
            raise RuntimeError("No data recorded")
        target_data = {}
        gen_data = {}
        for name in sorted(self._target_data):
            tensor = self._target_data[name]
            target_data[name] = tensor / self._n_batches[:, None, None]
            target_data[name] = self._dist.reduce_mean(target_data[name])
        for name in sorted(self._gen_data):
            tensor = self._gen_data[name]
            gen_data[name] = tensor / self._n_batches[:, None, None]
            gen_data[name] = self._dist.reduce_mean(gen_data[name])
        return gen_data, target_data


class _VarianceVideoData:
    """
    Record batches of video data and compute the variance.
    """

    def __init__(self, n_timesteps: int):
        self._target_means: TensorDict | None = None
        self._gen_means: TensorDict | None = None
        self._target_squares: TensorDict | None = None
        self._gen_squares: TensorDict | None = None
        self._n_timesteps = n_timesteps
        self._n_batches = torch.zeros([n_timesteps], dtype=torch.int32).cpu()
        self._dist = Distributed.get_instance()

    @torch.no_grad()
    def record_batch(
        self,
        target_data: TensorMapping,
        gen_data: TensorMapping,
        i_time_start: int,
    ):
        """
        Record a batch of data.

        Args:
            target_data: Dict of tensors of shape (n_samples, n_timesteps, ...)
            gen_data: Dict of tensors of shape (n_samples, n_timesteps, ...)
            i_time_start: Index of the first timestep in the batch.
        """
        if self._target_means is None:
            self._target_means = _initialize_video_from_batch(
                target_data, self._n_timesteps
            )
        if self._gen_means is None:
            self._gen_means = _initialize_video_from_batch(gen_data, self._n_timesteps)
        if self._target_squares is None:
            self._target_squares = _initialize_video_from_batch(
                target_data, self._n_timesteps
            )

        if self._gen_squares is None:
            self._gen_squares = _initialize_video_from_batch(
                gen_data, self._n_timesteps
            )

        window_steps = next(iter(target_data.values())).shape[1]
        time_slice = slice(i_time_start, i_time_start + window_steps)
        for name, tensor in target_data.items():
            self._target_means[name][time_slice, ...] += tensor.mean(dim=0).cpu()
            self._target_squares[name][time_slice, ...] += (tensor**2).mean(dim=0).cpu()
        for name, tensor in gen_data.items():
            self._gen_means[name][time_slice, ...] += tensor.mean(dim=0).cpu()
            self._gen_squares[name][time_slice, ...] += (tensor**2).mean(dim=0).cpu()
        self._n_batches[time_slice] += 1

    @torch.no_grad()
    def get(self) -> tuple[TensorDict, TensorDict]:
        if (
            self._gen_means is None
            or self._target_means is None
            or self._gen_squares is None
            or self._target_squares is None
        ):
            raise RuntimeError("No data recorded")
        target_data = {}
        gen_data = {}
        # calculate variance as E[X^2] - E[X]^2
        for name in sorted(self._target_means):
            tensor = self._target_means[name]
            mean = tensor / self._n_batches[:, None, None]
            mean = self._dist.reduce_mean(mean)
            square = self._target_squares[name] / self._n_batches[:, None, None]
            square = self._dist.reduce_mean(square)
            target_data[name] = square - mean**2
        for name in sorted(self._gen_means):
            tensor = self._gen_means[name]
            mean = tensor / self._n_batches[:, None, None]
            mean = self._dist.reduce_mean(mean)
            square = self._gen_squares[name] / self._n_batches[:, None, None]
            square = self._dist.reduce_mean(square)
            gen_data[name] = square - mean**2
        return gen_data, target_data


def _initialize_video_from_batch(
    batch: TensorMapping, n_timesteps: int, fill_value: float = 0.0
):
    """
    Initialize a video of the same shape as the batch, but with all valeus equal
    to fill_value and with n_timesteps timesteps.
    """
    video = {}
    for name, value in batch.items():
        shape = list(value.shape[1:])
        shape[0] = n_timesteps
        video[name] = torch.zeros(shape, dtype=torch.double).cpu()
        video[name][:, ...] = fill_value
    return video


@dataclasses.dataclass
class _MaybePairedVideoData:
    caption: str
    gen: torch.Tensor
    units: str | None
    long_name: str | None
    target: torch.Tensor | None = None

    def make_video(self):
        return _make_video(
            caption=self.caption,
            gen=self.gen,
            target=self.target,
        )


class VideoAggregator:
    """Videos of state evolution."""

    def __init__(
        self,
        n_timesteps: int,
        enable_extended_videos: bool,
        variable_metadata: Mapping[str, VariableMetadata] | None = None,
    ):
        """
        Args:
            n_timesteps: Number of timesteps of inference that will be run.
            enable_extended_videos: Whether to log videos of statistical
                metrics of state evolution
            variable_metadata: Mapping of variable names their metadata that will
                used in generating logged video captions.
        """
        if variable_metadata is None:
            self._variable_metadata: Mapping[str, VariableMetadata] = {}
        else:
            self._variable_metadata = variable_metadata
        self._mean_data = _MeanVideoData(n_timesteps=n_timesteps)
        if enable_extended_videos:
            self._error_data: _ErrorVideoData | None = _ErrorVideoData(
                n_timesteps=n_timesteps
            )
            self._variance_data: _VarianceVideoData | None = _VarianceVideoData(
                n_timesteps=n_timesteps
            )
            self._enable_extended_videos = True
        else:
            self._error_data = None
            self._variance_data = None
            self._enable_extended_videos = False

    @torch.no_grad()
    def record_batch(
        self,
        target_data: TensorMapping,
        gen_data: TensorMapping,
        target_data_norm: TensorMapping | None = None,
        gen_data_norm: TensorMapping | None = None,
        i_time_start: int = 0,
    ):
        del target_data_norm, gen_data_norm  # intentionally unused
        self._mean_data.record_batch(
            target_data=target_data,
            gen_data=gen_data,
            i_time_start=i_time_start,
        )
        if self._error_data is not None:
            self._error_data.record_batch(
                target_data=target_data,
                gen_data=gen_data,
                i_time_start=i_time_start,
            )
        if self._variance_data is not None:
            self._variance_data.record_batch(
                target_data=target_data,
                gen_data=gen_data,
                i_time_start=i_time_start,
            )

    @torch.no_grad()
    def get_logs(self, label: str):
        """
        Returns logs as can be reported to WandB.

        Args:
            label: Label to prepend to all log keys.
        """
        data = self._get_data()
        videos = {}
        for sub_label, d in data.items():
            videos[f"{label}/{sub_label}"] = d.make_video()
        return videos

    @torch.no_grad()
    def _get_data(self) -> Mapping[str, _MaybePairedVideoData]:
        """
        Returns video data as can be reported to WandB.

        Args:
            label: Label to prepend to all log keys.
        """
        gen_data, target_data = self._mean_data.get()
        video_data = {}

        def get_units(name: str) -> str | None:
            if name in self._variable_metadata:
                return self._variable_metadata[name].units
            else:
                return None

        def get_long_name(name: str) -> str | None:
            if name in self._variable_metadata:
                return self._variable_metadata[name].long_name
            else:
                return None

        for name in gen_data:
            video_data[name] = _MaybePairedVideoData(
                caption=self._get_caption(name),
                gen=gen_data[name],
                target=target_data[name],
                units=get_units(name),
                long_name=f"ensemble mean of {get_long_name(name)}",
            )
            if self._enable_extended_videos:
                video_data[f"bias/{name}"] = _MaybePairedVideoData(
                    caption=(f"prediction - target for {name}"),
                    gen=gen_data[name] - target_data[name],
                    units=get_units(name),
                    long_name=f"bias of {get_long_name(name)}",
                )
        if self._error_data is not None:
            data = self._error_data.get()
            for name in data.rmse:
                video_data[f"rmse/{name}"] = _MaybePairedVideoData(
                    caption=f"RMSE over ensemble for {name}",
                    gen=data.rmse[name],
                    units=get_units(name),
                    long_name=f"root mean squared error of {get_long_name(name)}",
                )
            for name in data.min_err:
                video_data[f"min_err/{name}"] = _MaybePairedVideoData(
                    caption=f"Min across ensemble members of min error for {name}",
                    gen=data.min_err[name],
                    units=get_units(name),
                    long_name=(
                        f"min error of {get_long_name(name)} across ensemble members"
                    ),
                )
            for name in data.max_err:
                video_data[f"max_err/{name}"] = _MaybePairedVideoData(
                    caption=f"Max across ensemble members of max error for {name}",
                    gen=data.max_err[name],
                    units=get_units(name),
                    long_name=(
                        f"max error of {get_long_name(name)} across ensemble members"
                    ),
                )
        if self._variance_data is not None:
            gen_data, target_data = self._variance_data.get()
            for name in gen_data:
                video_data[f"gen_var/{name}"] = _MaybePairedVideoData(
                    caption=(
                        f"Variance of gen data for {name} "
                        "as fraction of target variance"
                    ),
                    gen=gen_data[name] / target_data[name],
                    units="",
                    long_name=(
                        f"prediction variance of {get_long_name(name)} "
                        "as fraction of target variance"
                    ),
                )
        return video_data

    @torch.no_grad()
    def get_dataset(self) -> xr.Dataset:
        """
        Return video data as an xarray Dataset.
        """
        data = self._get_data()
        video_data = {}
        for label, d in data.items():
            label = label.strip("/").replace("/", "_")  # remove leading slash
            attrs = {}
            if d.units is not None:
                attrs["units"] = d.units
            if d.long_name is not None:
                attrs["long_name"] = d.long_name
            if d.target is not None:
                video_data[label] = xr.DataArray(
                    data=np.concatenate(
                        [d.gen.cpu().numpy()[None, :], d.target.cpu().numpy()[None, :]],
                        axis=0,
                    ),
                    dims=("source", "timestep", "lat", "lon"),
                    attrs=attrs,
                )
            else:
                video_data[label] = xr.DataArray(
                    data=d.gen.cpu().numpy(),
                    dims=("timestep", "lat", "lon"),
                    attrs=attrs,
                )
        return xr.Dataset(video_data)

    def _get_caption(self, name: str) -> str:
        caption = (
            "Autoregressive (left) prediction and (right) target for {name} [{units}]"
        )
        if name in self._variable_metadata:
            caption_name = self._variable_metadata[name].long_name
            units = self._variable_metadata[name].units
        else:
            caption_name, units = name, "unknown units"
        return caption.format(name=caption_name, units=units)


def _make_video(
    caption: str,
    gen: torch.Tensor,
    target: torch.Tensor | None = None,
):
    if target is None:
        video_data = np.expand_dims(gen.cpu().numpy(), axis=1)
    else:
        gen = np.expand_dims(gen.cpu().numpy(), axis=1)
        target = np.expand_dims(target.cpu().numpy(), axis=1)
        gap = np.zeros([gen.shape[0], 1, gen.shape[2], 10])
        video_data = np.concatenate([gen, gap, target], axis=-1)
    if target is None:
        data_min = np.nanmin(video_data)
        data_max = np.nanmax(video_data)
    else:
        # use target data to set the color scale
        data_min = np.nanmin(target)
        data_max = np.nanmax(target)
    # video data is brightness values on a 0-255 scale
    video_data = 255 * (video_data - data_min) / (data_max - data_min)
    video_data = np.minimum(video_data, 255)
    video_data = np.maximum(video_data, 0)
    video_data[np.isnan(video_data)] = 0
    caption += f"; vmin={data_min:.4g}, vmax={data_max:.4g}"
    return wandb.Video(
        np.flip(video_data, axis=-2),
        caption=caption,
        fps=4,
    )



"""Integration test for inference entrypoint."""

import dataclasses
import datetime
import pathlib

import cftime
import numpy as np
import pytest
import torch
import xarray as xr
import yaml

from fme.ace.data_loading.batch_data import PrognosticState
from fme.ace.data_loading.inference import (
    ExplicitIndices,
    ForcingDataLoaderConfig,
    InferenceInitialConditionIndices,
    TimestampList,
)
from fme.ace.inference.data_writer import DataWriterConfig
from fme.ace.inference.inference import (
    InferenceConfig,
    InitialConditionConfig,
    get_initial_condition,
    main,
)
from fme.ace.registry import ModuleSelector
from fme.ace.stepper import SingleModuleStepperConfig
from fme.ace.testing import DimSizes, FV3GFSData
from fme.core.coordinates import (
    DimSize,
    HybridSigmaPressureCoordinate,
    LatLonCoordinates,
)
from fme.core.dataset.data_typing import VariableMetadata
from fme.core.dataset_info import DatasetInfo
from fme.core.logging_utils import LoggingConfig
from fme.core.normalizer import NormalizationConfig
from fme.core.ocean import OceanConfig
from fme.core.testing import mock_wandb

TIMESTEP = datetime.timedelta(hours=6)


class PlusOne(torch.nn.Module):
    def forward(self, x):
        return x + 1


def save_stepper(
    path: pathlib.Path,
    in_names: list[str],
    out_names: list[str],
    mean: float,
    std: float,
    horizontal_coords: dict[str, xr.DataArray],
    nz_interface: int,
    timestep: datetime.timedelta = TIMESTEP,
):
    all_names = list(set(in_names).union(out_names))
    config = SingleModuleStepperConfig(
        builder=ModuleSelector(type="prebuilt", config={"module": PlusOne()}),
        in_names=in_names,
        out_names=out_names,
        normalization=NormalizationConfig(
            means={name: mean for name in all_names},
            stds={name: std for name in all_names},
        ),
        ocean=OceanConfig(
            surface_temperature_name="sst",
            ocean_fraction_name="ocean_fraction",
        ),
    )
    horizontal_coordinate = LatLonCoordinates(
        lat=torch.tensor(horizontal_coords["lat"].values, dtype=torch.float32),
        lon=torch.tensor(horizontal_coords["lon"].values, dtype=torch.float32),
    )
    vertical_coordinate = HybridSigmaPressureCoordinate(
        ak=torch.arange(nz_interface), bk=torch.arange(nz_interface)
    )
    dataset_info = DatasetInfo(
        horizontal_coordinates=horizontal_coordinate,
        vertical_coordinate=vertical_coordinate,
        timestep=timestep,
        variable_metadata={
            "prog": VariableMetadata(
                units="m",
                long_name="a prognostic variable",
            ),
        },
    )
    stepper = config.get_stepper(
        dataset_info=dataset_info,
    )
    torch.save({"stepper": stepper.get_state()}, path)


def test_inference_entrypoint(tmp_path: pathlib.Path):
    forward_steps_in_memory = 2
    # NOTE: number of inputs and outputs has to be the same for the PlusOne
    # stepper module to work properly
    in_names = ["prog", "sst", "forcing_var", "DSWRFtoa"]
    out_names = ["prog", "sst", "ULWRFtoa", "USWRFtoa"]
    stepper_path = tmp_path / "stepper"
    horizontal = [DimSize("lat", 16), DimSize("lon", 32)]
    nz_interface = 4

    dim_sizes = DimSizes(
        n_time=9,
        horizontal=horizontal,
        nz_interface=4,
    )
    data = FV3GFSData(
        path=tmp_path,
        names=["forcing_var", "DSWRFtoa", "sst", "ocean_fraction"],
        dim_sizes=dim_sizes,
        timestep_days=0.25,
        save_vertical_coordinate=False,
    )
    save_stepper(
        stepper_path,
        in_names=in_names,
        out_names=out_names,
        mean=0.0,
        std=1.0,
        horizontal_coords=data.horizontal_coords,
        nz_interface=nz_interface,
    )
    dims = ["sample", "lat", "lon"]
    initial_condition = xr.Dataset(
        {
            "prog": xr.DataArray(
                np.random.rand(2, 16, 32).astype(np.float32), dims=dims
            ),
            "sst": xr.DataArray(
                np.random.rand(2, 16, 32).astype(np.float32), dims=dims
            ),
            "DSWRFtoa": xr.DataArray(
                np.random.rand(2, 16, 32).astype(np.float32), dims=dims
            ),
        }
    )

    initial_condition_path = tmp_path / "init_data" / "ic.nc"
    initial_condition_path.parent.mkdir()
    initial_condition["time"] = xr.DataArray(
        [
            cftime.DatetimeProlepticGregorian(2000, 1, 1, 6),
            cftime.DatetimeProlepticGregorian(2000, 1, 1, 18),
        ],
        dims=["time"],
    )
    initial_condition.to_netcdf(initial_condition_path, mode="w")
    forcing_loader = ForcingDataLoaderConfig(
        dataset=data.inference_data_loader_config.dataset,
        num_data_workers=data.inference_data_loader_config.num_data_workers,
    )

    config = InferenceConfig(
        experiment_dir=str(tmp_path),
        n_forward_steps=4,
        forward_steps_in_memory=forward_steps_in_memory,
        checkpoint_path=str(stepper_path),
        logging=LoggingConfig(
            log_to_screen=True,
            log_to_file=False,
            log_to_wandb=True,
        ),
        initial_condition=InitialConditionConfig(path=str(initial_condition_path)),
        forcing_loader=forcing_loader,
        data_writer=DataWriterConfig(save_prediction_files=True),
        allow_incompatible_dataset=False,
    )
    config_filename = tmp_path / "config.yaml"
    with open(config_filename, "w") as f:
        yaml.dump(dataclasses.asdict(config), f)

    with mock_wandb() as wandb:
        wandb.configure(log_to_wandb=True)
        main(yaml_config=str(config_filename))
        wandb_logs = wandb.get_logs()

    n_ic_timesteps = 1
    summary_log_step = 1
    assert len(wandb_logs) == n_ic_timesteps + config.n_forward_steps + summary_log_step
    for i, log in enumerate(wandb_logs):
        for metric, val in log.items():
            # check that time series metrics match
            if "inference/mean" in metric:
                if i > 0:
                    assert metric in wandb_logs[i]
                    if np.isnan(val):
                        assert np.isnan(wandb_logs[i][metric])
                    else:
                        assert wandb_logs[i][metric] == val
                elif not np.isnan(val):  # for IC only valid data is reported to wandb
                    assert metric in wandb_logs[i]
                    assert wandb_logs[i][metric] == val

    ds = xr.open_dataset(
        tmp_path / "autoregressive_predictions.nc", decode_timedelta=False
    )
    # prognostic in
    assert "prog" in ds
    assert ds["prog"].attrs == {  # this should come from the stepper metadata
        "units": "m",
        "long_name": "a prognostic variable",
    }
    # diags in
    assert "ULWRFtoa" in ds
    assert ds["ULWRFtoa"].attrs == {  # this should come from default metadata
        "units": "W/m**2",
        "long_name": "Upward LW radiative flux at TOA",
    }
    assert "USWRFtoa" in ds

    # derived in
    assert "net_energy_flux_toa_into_atmosphere" in ds
    assert "units" in ds.net_energy_flux_toa_into_atmosphere.attrs
    assert "long_name" in ds.net_energy_flux_toa_into_atmosphere.attrs
    # forcings not in
    assert "DSWRFtoa" not in ds
    assert "forcing_var" not in ds
    assert ds["prog"].sizes == {"time": 4, "sample": 2, "lat": 16, "lon": 32}
    np.testing.assert_allclose(
        ds["prog"].isel(time=0).values, initial_condition["prog"].values + 1
    )
    np.testing.assert_allclose(
        ds["prog"].isel(time=1).values, ds["prog"].isel(time=0).values + 1, rtol=1e-6
    )
    saved_data = xr.open_dataset(data.data_filename, decode_timedelta=False)
    ops = LatLonCoordinates(
        lat=torch.as_tensor(saved_data["lat"].values.astype(np.float32)),
        lon=torch.as_tensor(saved_data["lon"].values.astype(np.float32)),
    ).get_gridded_operations()
    # check that inference logs match raw output
    for i in range(1, config.n_forward_steps + 1):
        for log_name in wandb_logs[i]:
            if "inference/mean/weighted_mean_gen" in log_name:
                variable_name = log_name.split("/")[-1]
                # note raw output does not include initial condition, hence
                # i-1 below. Code uses area from data, not stepper above.
                raw_variable = ds[variable_name].isel(time=i - 1)
                raw_global_mean = ops.area_weighted_mean(
                    torch.as_tensor(raw_variable.values)
                ).mean()
                np.testing.assert_allclose(
                    raw_global_mean, wandb_logs[i][log_name], rtol=1e-6
                )
    assert "inference/total_steps_per_second" in wandb_logs[-1]


def test_get_initial_condition():
    time_da = xr.DataArray([0, 5], dims=["sample"])
    prognostic_da = xr.DataArray(
        np.random.rand(2, 16, 32), dims=["sample", "lat", "lon"]
    )
    data = xr.Dataset({"prog": prognostic_da, "time": time_da})
    initial_condition = get_initial_condition(data, ["prog"])
    assert isinstance(initial_condition, PrognosticState)
    batch_data = initial_condition.as_batch_data()
    assert batch_data.time.shape == (2, 1)
    initial_times = batch_data.time.isel(time=0)
    assert initial_times.shape == (2,)
    assert initial_times[0] == 0
    assert initial_times[1] == 5
    assert batch_data.data["prog"].shape == (2, 1, 16, 32)
    np.testing.assert_allclose(
        batch_data.data["prog"].squeeze(dim=1).cpu().numpy(), data["prog"].values
    )
    assert batch_data.time.isel(time=0).equals(initial_times)


def test_get_initial_condition_raises_bad_variable_shape():
    time_da = xr.DataArray([0, 5], dims=["sample"])
    prognostic_da = xr.DataArray(
        np.random.rand(2, 8, 16, 32), dims=["sample", "layer", "lat", "lon"]
    )
    data = xr.Dataset({"prog": prognostic_da, "time": time_da})
    with pytest.raises(ValueError):
        get_initial_condition(data, ["prog"])


def test_get_initial_condition_raises_missing_time():
    prognostic_da = xr.DataArray(
        np.random.rand(2, 16, 32), dims=["sample", "lat", "lon"]
    )
    data = xr.Dataset({"prog": prognostic_da})
    with pytest.raises(ValueError):
        get_initial_condition(data, ["prog"])


def test_get_initial_condition_raises_mismatched_time_length():
    time_da = xr.DataArray([0, 5, 10], dims=["time"])
    prognostic_da = xr.DataArray(
        np.random.rand(2, 16, 32), dims=["sample", "lat", "lon"]
    )
    data = xr.Dataset({"prog": prognostic_da, "time": time_da})
    with pytest.raises(ValueError):
        get_initial_condition(data, ["prog"])


@pytest.mark.parametrize(
    "start_indices, expected_time_values",
    [
        (
            None,
            [
                cftime.datetime(2000, 1, 1, 6),
                cftime.datetime(2000, 1, 1, 12),
                cftime.datetime(2000, 1, 1, 18),
            ],
        ),
        (
            InferenceInitialConditionIndices(
                n_initial_conditions=2, first=0, interval=2
            ),
            [cftime.datetime(2000, 1, 1, 6), cftime.datetime(2000, 1, 1, 18)],
        ),
        (
            ExplicitIndices([0, 2]),
            [cftime.datetime(2000, 1, 1, 6), cftime.datetime(2000, 1, 1, 18)],
        ),
        (
            TimestampList(times=["2000-01-01T06:00:00", "2000-01-01T18:00:00"]),
            [cftime.datetime(2000, 1, 1, 6), cftime.datetime(2000, 1, 1, 18)],
        ),
        (
            ExplicitIndices([1]),
            [
                cftime.datetime(2000, 1, 1, 12),
            ],
        ),
    ],
)
def test__subselect_initial_conditions(tmp_path, start_indices, expected_time_values):
    initial_condition = xr.Dataset(
        {"prog": xr.DataArray(np.random.rand(3, 16, 32), dims=["sample", "lat", "lon"])}
    )
    initial_condition_path = tmp_path / "init_data" / "ic.nc"
    initial_condition_path.parent.mkdir()
    initial_condition["time"] = xr.DataArray(
        [
            cftime.datetime(2000, 1, 1, 6),
            cftime.datetime(2000, 1, 1, 12),
            cftime.datetime(2000, 1, 1, 18),
        ],
        dims=["sample"],
    )
    initial_condition.to_netcdf(initial_condition_path, mode="w")

    ic_config = InitialConditionConfig(
        path=str(initial_condition_path),
        start_indices=start_indices,
    )
    ic_data = ic_config.get_dataset()

    np.testing.assert_array_equal(ic_data.time.values, np.array(expected_time_values))
    assert ic_data["prog"].shape == (len(expected_time_values), 16, 32)



import torch


def compare_restored_parameters(module1_params, module2_params, optimizer1, optimizer2):
    """
    This helper function is used to compare the model parameters and optimizer
    states after restoring from a checkpoint.

    This covered a tricky failure case where the optimizer state was not
    restored correctly, preventing the restored model from training any further.
    https://github.com/ai2cm/full-model/pull/1936

    Note that `optimizer.state` objects are defaultdicts where keys are the model
    parameter tensors and the values are optimizer state dictionaries.
    """
    for param1, param2 in zip(
        module1_params,
        module2_params,
    ):
        assert torch.equal(param1, param2)
        assert param1 in optimizer1.state
        assert param2 in optimizer2.state

        # optimizer state is a defaultdict with the parameter tensor as the key
        # this links the model to relevant optimizer state for training
        # https://docs.pytorch.org/docs/stable/_modules/torch/optim/optimizer.html#Optimizer.load_state_dict # noqa: E501
        optimizer_state1 = optimizer1.state[param1]
        optimizer_state2 = optimizer2.state[param2]

        for key, value1 in optimizer_state1.items():
            assert key in optimizer_state2
            value2 = optimizer_state2[key]
            if key == "step":
                # step is not put on device, but state restore loads to GPU
                value2 = value2.cpu()
            assert torch.equal(value1, value2)



import datetime

import torch

from fme.core.ocean import (
    Ocean,
    OceanConfig,
    SlabOceanConfig,
    mixed_layer_temperature_tendency,
)

TIMESTEP = datetime.timedelta(hours=6)


def test_ocean_prescribed():
    config = OceanConfig(surface_temperature_name="sst", ocean_fraction_name="of")
    ocean = Ocean(config, timestep=TIMESTEP)
    target_data = {"sst": torch.tensor([22.0, 25.0]), "of": torch.tensor([0.2, 0.8])}
    input_data = {"sst": torch.tensor([20.0, 21.0]), "foo": torch.tensor([1, 2])}
    gen_data = {"sst": torch.tensor([23.0, 26.0]), "foo": torch.tensor([2, 3])}
    output_data = ocean(input_data, gen_data, target_data)
    expected_output = {"sst": torch.tensor([23.0, 25.0]), "foo": torch.tensor([2, 3])}
    assert set(output_data) == set(expected_output)
    for name in output_data:
        torch.testing.assert_close(output_data[name], expected_output[name])


def test_ocean_slab():
    config = OceanConfig(
        surface_temperature_name="sst",
        ocean_fraction_name="of",
        slab=SlabOceanConfig(
            mixed_layer_depth_name="mld",
            q_flux_name="qf",
        ),
    )
    names_for_net_surface_energy_flux = [
        "DLWRFsfc",
        "ULWRFsfc",
        "DSWRFsfc",
        "USWRFsfc",
        "LHTFLsfc",
        "SHTFLsfc",
    ]
    fluxes = {k: torch.tensor([2.0]) for k in names_for_net_surface_energy_flux}
    expected_net_surface_energy_flux = torch.tensor([-4.0])
    ocean = Ocean(config, timestep=TIMESTEP)
    target_data = {
        "mld": torch.tensor([25.0]),
        "of": torch.tensor([0.8]),
        "qf": torch.tensor([40.0]),
    }
    input_data = {"sst": torch.tensor([20.0])}
    gen_data = {**fluxes, "sst": torch.tensor([25.0])}
    output_data = ocean(input_data, gen_data, target_data)
    expected_sst_tendency = mixed_layer_temperature_tendency(
        expected_net_surface_energy_flux, target_data["qf"], target_data["mld"]
    )
    timestep_seconds = TIMESTEP / datetime.timedelta(seconds=1)
    expected_sst = input_data["sst"] + timestep_seconds * expected_sst_tendency
    expected_output = {**fluxes, "sst": expected_sst}
    assert set(output_data) == set(expected_output)
    for name in output_data:
        torch.testing.assert_close(output_data[name], expected_output[name])


def test_mixed_layer_temperature_tendency():
    f_net = torch.tensor([10.0])
    q_flux = torch.tensor([5.0])
    depth = torch.tensor([100.0])
    result = mixed_layer_temperature_tendency(
        f_net, q_flux, depth, density=5.0, specific_heat=3.0
    )
    expected_result = (f_net + q_flux) / (5 * 3 * depth)
    torch.testing.assert_close(result, expected_result)



import dataclasses

from fme.core.masking import replace_on_mask
from fme.core.typing_ import TensorDict, TensorMapping


@dataclasses.dataclass
class PrescriberConfig:
    """
    Configuration for overwriting predictions of 'prescribed_name' by target values.

    If interpolate is False, the data is overwritten in the region where
    'mask_name' == 'mask_value' after values are rounded to integer. If interpolate
    is True, the data is interpolated between the predicted value at 0 and the
    target value at 1 based on the mask variable, and it is assumed the mask variable
    lies in the range from 0 to 1.

    Parameters:
        prescribed_name: Name of the variable to be overwritten.
        mask_name: Name of the mask variable.
        mask_value: Value of the mask variable in the region to be overwritten.
        interpolate: Whether to interpolate linearly between the generated and target
            values in the masked region, where 0 means keep the generated values and
            1 means replace completely with the target values. Requires mask_value
            be set to 1.
    """

    prescribed_name: str
    mask_name: str
    mask_value: int
    interpolate: bool = False

    def __post_init__(self):
        if self.interpolate and self.mask_value != 1:
            raise ValueError(
                "Interpolation requires mask_value to be 1, but it is set to "
                f"{self.mask_value}."
            )

    def build(self, in_names: list[str], out_names: list[str]):
        if not (self.prescribed_name in in_names and self.prescribed_name in out_names):
            raise ValueError(
                "Variables which are being prescribed in masked regions must be in"
                f" in_names and out_names, but {self.prescribed_name} is not."
            )
        return Prescriber(
            prescribed_name=self.prescribed_name,
            mask_name=self.mask_name,
            mask_value=self.mask_value,
            interpolate=self.interpolate,
        )


class Prescriber:
    """
    Class to overwrite 'prescribed_name' by target values in masked regions.
    """

    def __init__(
        self,
        prescribed_name: str,
        mask_name: str,
        mask_value: int,
        interpolate: bool = False,
    ):
        self.prescribed_name = prescribed_name
        self.mask_name = mask_name
        self.mask_value = mask_value
        self.interpolate = interpolate

    def __call__(
        self,
        mask_data: TensorMapping,
        gen: TensorMapping,
        target: TensorMapping,
    ) -> TensorDict:
        """
        Args:
            mask_data: Dictionary of data containing the mask variable.
            gen: Dictionary of data to use outside of mask region.
            target: Dictionary of data to use in mask region.

        Returns:
            Dictionary of data with the prescribed variable overwritten in the mask
            region and other variables unmodified from gen.
        """
        for name, named_tensors in [("gen", gen), ("target", target)]:
            if self.prescribed_name not in named_tensors:
                raise ValueError(
                    f'Prescribed variable "{self.prescribed_name}" '
                    f'is missing from "{name}"'
                )

        if self.interpolate:
            mask = mask_data[self.mask_name]
            # 1 keeps the target values, 0 replaces with the gen values
            output = (
                mask * target[self.prescribed_name]
                + (1 - mask) * gen[self.prescribed_name]
            )
        else:
            # overwrite specified target variable in given mask region
            output = replace_on_mask(
                original=gen[self.prescribed_name],
                replacement=target[self.prescribed_name],
                mask=mask_data[self.mask_name],
                mask_value=self.mask_value,
            )
        return {**gen, self.prescribed_name: output}

    @property
    def prescribed_names(self) -> list[str]:
        return [self.prescribed_name]

    @property
    def mask_names(self) -> list[str]:
        return [self.mask_name]



import contextlib
import logging
import time
import warnings

import numpy as np

INACTIVE_WARNING_MESSAGE = (
    "The GlobalTimer is currently inactive; therefore no timing information "
    "will be recorded. To activate it, wrap your code within the GlobalTimer() "
    "context."
)


class CumulativeTimer:
    def __init__(self, category):
        self._duration = 0.0
        self._start_time = None
        self._category = category

    def start(self):
        if self._start_time is not None:
            raise RuntimeError(f"timer {self._category!r} is already running")
        self._start_time = time.time()

    def stop(self):
        if self._start_time is None:
            raise RuntimeError(
                f"must call start for timer {self._category!r} before stop"
            )
        self._duration += time.time() - self._start_time
        self._start_time = None

    @property
    def duration(self) -> float:
        if self._start_time is not None:
            raise RuntimeError(f"timer {self._category!r} is still running")
        return self._duration


class GlobalTimer:
    """
    A singleton class to make timing inference code easier.
    """

    @classmethod
    def get_instance(cls) -> "GlobalTimer":
        """
        Get the singleton instance of the GlobalTimer class.
        """
        global singleton
        if singleton is None:
            singleton = cls()
            warnings.warn(INACTIVE_WARNING_MESSAGE)
        return singleton

    @classmethod
    def __enter__(cls):
        global singleton
        if singleton is None:
            singleton = cls()
        if singleton._active:
            raise RuntimeError("GlobalTimer is currently in use in another context")
        singleton._active = True

    @classmethod
    def __exit__(cls, type, value, traceback):
        global singleton
        singleton = None

    def __init__(self):
        self._timers: dict[str, CumulativeTimer] = {}
        self._active = False
        self._current_category: str | None = None

    def outer_context(self, category: str) -> contextlib.AbstractContextManager:
        """
        Context manager for timing a block of code.

        May be active at the same time as other timers.
        """

        @contextlib.contextmanager
        def timer_context():
            self.start_outer(category)
            try:
                yield
            finally:
                self.stop_outer(category)

        return timer_context()

    def context(self, category: str) -> contextlib.AbstractContextManager:
        """
        Context manager for timing a block of code.

        Only one inner timer can be active at a time.
        """

        @contextlib.contextmanager
        def timer_context():
            self.start(category)
            try:
                yield
            finally:
                self.stop()

        return timer_context()

    def start(self, category: str):
        """
        Start an inner timer for the given category.

        Only one inner timer can be active at a time.
        """
        if self._current_category is not None:
            raise RuntimeError(
                "GlobalTimer already has an active inner timer, "
                f"{self._current_category}"
            )
        self.start_outer(category)
        self._current_category = category

    def start_outer(self, category: str):
        """
        Start a timer for the given category.

        May be active at the same time as other timers.
        """
        if self._active:
            if category not in self._timers:
                self._timers[category] = CumulativeTimer(category)
            self._timers[category].start()

    def stop(self):
        """
        Stop the currently active inner timer.
        """
        if self._current_category is None:
            raise RuntimeError("GlobalTimer does not have a running timer")
        self.stop_outer(self._current_category)
        self._current_category = None

    def stop_outer(self, category: str):
        """
        Stop the timer for the given category.

        Does not change the currently active inner timer.
        """
        if self._active:
            self._timers[category].stop()

    def get_duration(self, category: str) -> float:
        if self._active:
            return self._timers[category].duration
        else:
            return np.nan

    def get_durations(self) -> dict[str, float]:
        return {category: timer.duration for category, timer in self._timers.items()}

    def log_durations(self):
        for name, duration in self.get_durations().items():
            logging.info(f"{name} duration: {duration:.2f}s")


singleton: GlobalTimer | None = None



import argparse
import dataclasses
import logging
import os
from typing import List, Sequence, Tuple

import dacite
import torch.utils.data
import xarray as xr
import yaml

import fme.core.logging_utils as logging_utils
from fme.ace.data_loading.batch_data import BatchData, default_collate
from fme.ace.data_loading.config import (
    ConcatDatasetConfig,
    DataLoaderConfig,
    MergeDatasetConfig,
)
from fme.ace.inference.data_writer.dataset_metadata import DatasetMetadata
from fme.ace.inference.data_writer.monthly import (
    MonthlyDataWriter,
    months_for_timesteps,
)
from fme.ace.requirements import DataRequirements
from fme.core.coordinates import (
    AtmosphericDeriveFn,
    OptionalHybridSigmaPressureCoordinate,
)
from fme.core.dataset.getters import get_datasets, get_merged_datasets
from fme.core.dataset.properties import DatasetProperties
from fme.core.device import using_gpu
from fme.core.distributed import Distributed
from fme.core.logging_utils import LoggingConfig
from fme.core.typing_ import TensorMapping


@dataclasses.dataclass
class CollateFn:
    horizontal_dims: List[str]

    def __call__(
        self, samples: Sequence[Tuple[TensorMapping, xr.DataArray]]
    ) -> "BatchData":
        sample_data, sample_time = zip(*samples)
        batch_data = default_collate(sample_data)
        batch_time = xr.concat(sample_time, dim="sample")
        return BatchData(
            data=batch_data,
            time=batch_time,
            horizontal_dims=self.horizontal_dims,
        )


def get_data_loaders(
    config: DataLoaderConfig, requirements: DataRequirements
) -> Tuple[List[torch.utils.data.DataLoader], DatasetProperties]:
    dist = Distributed.get_instance()
    if dist.world_size > 1:
        raise RuntimeError(
            "Data loading for write_monthly_data.py is not "
            "supported in distributed mode."
        )
    datasets: torch.utils.data.Dataset
    if isinstance(config.dataset, ConcatDatasetConfig):
        datasets, properties = get_datasets(
            config.dataset.concat, requirements.names, requirements.n_timesteps
        )
    elif isinstance(config.dataset, MergeDatasetConfig):
        datasets, properties = get_merged_datasets(
            config.dataset,
            requirements.names,
            requirements.n_timesteps,
        )

    data_loaders = []
    for dataset in datasets:
        dataloader = torch.utils.data.DataLoader(
            dataset,
            batch_size=config.batch_size,
            num_workers=config.num_data_workers,
            shuffle=False,
            sampler=None,
            drop_last=True,
            pin_memory=using_gpu(),
            collate_fn=CollateFn(
                horizontal_dims=list(properties.horizontal_coordinates.dims),
            ),
        )
        data_loaders.append(dataloader)
    return data_loaders, properties


def get_timesteps(data_loaders: List[torch.utils.data.DataLoader]) -> int:
    timesteps_per_dataset = len(data_loaders[0].dataset)
    if not all(len(loader.dataset) == timesteps_per_dataset for loader in data_loaders):
        raise ValueError("All datasets must have the same number of timesteps.")
    return timesteps_per_dataset


@dataclasses.dataclass
class Config:
    """
    Configuration for applying the MonthlyDataWriter to a dataset.

    Parameters:
        experiment_dir: Directory to save results to.
        dataset: Configuration for the dataset to load.
        num_data_workers: Number of parallel workers to use for data loading.
        logging: Configuration for logging.
        variable_names: Names of the variables to write to disk.
    """

    experiment_dir: str
    data_loader: DataLoaderConfig
    variable_names: List[str]
    logging: LoggingConfig = dataclasses.field(
        default_factory=lambda: LoggingConfig(
            log_to_file=True, log_to_screen=False, log_to_wandb=False
        )
    )

    def __post_init__(self):
        if self.data_loader.batch_size != 1:
            raise ValueError("Batch size must be 1 to write dataset using writer.")

    def get_data(self) -> "Data":
        data_loaders, properties = get_data_loaders(
            config=self.data_loader,
            requirements=DataRequirements(
                names=self.variable_names,
                n_timesteps=1,  # 1 to count steps properly, avoid window overlap
            ),
        )
        n_timesteps = get_timesteps(data_loaders=data_loaders)
        return Data(
            loaders=data_loaders,
            properties=properties,
            n_timesteps=n_timesteps,
        )

    def configure_logging(self, log_filename: str):
        self.logging.configure_logging(self.experiment_dir, log_filename)

    def get_data_writer(self, data: "Data") -> MonthlyDataWriter:
        n_months = months_for_timesteps(data.n_timesteps, data.properties.timestep)
        coords = {
            **data.properties.horizontal_coordinates.coords,
            **data.properties.vertical_coordinate.coords,
        }
        return MonthlyDataWriter(
            path=self.experiment_dir,
            label="data",
            save_names=None,  # save all data given
            n_samples=self.data_loader.batch_size * len(data.loaders),
            n_months=n_months,
            variable_metadata=data.properties.variable_metadata,
            coords=coords,
            dataset_metadata=DatasetMetadata.from_env(),
        )


@dataclasses.dataclass
class Data:
    loaders: List[torch.utils.data.DataLoader]
    properties: DatasetProperties
    n_timesteps: int


def merge_loaders(loaders: List[torch.utils.data.DataLoader]):
    for window_batch_data_list in zip(*loaders):
        tensors = [item.data for item in window_batch_data_list]
        time = [item.time for item in window_batch_data_list]
        window_batch_data = {
            k: torch.concat([d[k] for d in tensors]) for k in tensors[0].keys()
        }
        time = xr.concat(time, dim="sample")
        yield BatchData(
            data=window_batch_data,
            time=time,
        )


def run(config: Config):
    config.configure_logging(log_filename="write_monthly_data_out.log")
    logging_utils.log_versions()

    data = config.get_data()
    writer = config.get_data_writer(data)

    assert isinstance(
        data.properties.vertical_coordinate, OptionalHybridSigmaPressureCoordinate
    )
    derive_func = AtmosphericDeriveFn(
        vertical_coordinate=data.properties.vertical_coordinate,
        timestep=data.properties.timestep,
    )

    n_batches = len(data.loaders[0].dataset) // config.data_loader.batch_size
    for i, window_batch_data in enumerate(merge_loaders(data.loaders)):
        # no need to trim initial conditions because
        # we set n_timesteps to 1 in the DataRequirements
        assert list(window_batch_data.data.values())[0].shape[1] == 1

        window_batch_data = window_batch_data.compute_derived_variables(
            derive_func=derive_func,
            forcing_data=window_batch_data,
        )
        writer.append_batch(
            data=window_batch_data.data,
            start_timestep=-1,  # ignored
            batch_time=window_batch_data.time,
        )
        if i % 10 == 0:
            logging.info(f"Writing batch {i + 1} of {n_batches}.")
            writer.flush()

    writer.flush()
    logging.info("Finished writing data.")


def main(
    yaml_config: str,
):
    with open(yaml_config, "r") as f:
        data = yaml.safe_load(f)
    config = dacite.from_dict(
        data_class=Config,
        data=data,
        config=dacite.Config(strict=True),
    )
    if not os.path.isdir(config.experiment_dir):
        os.makedirs(config.experiment_dir)
    with open(os.path.join(config.experiment_dir, "config.yaml"), "w") as f:
        yaml.dump(data, f, default_flow_style=False, sort_keys=False)
    run(config)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("yaml_config", type=str)

    args = parser.parse_args()

    main(
        yaml_config=args.yaml_config,
    )



import os

import pytest
import torch
import torch.multiprocessing as mp

from fme import get_device

from .distributed import Distributed, pad_tensor_at_end, unpad_tensor_at_end


@pytest.mark.parametrize(
    ["padding", "fill_value"],
    [
        pytest.param([0, 0, 0], None, id="no_padding"),
        pytest.param([1, 1, 1], 0.0, id="padding_1"),
        pytest.param([1, 1, 1], 1.0, id="padding_1_fill_one"),
    ],
)
def test_pad_tensor_at_end(padding, fill_value):
    tensor = torch.ones(2, 3, 4)
    padded_tensor = pad_tensor_at_end(tensor, padding, fill_value)
    assert padded_tensor.size() == (2 + padding[0], 3 + padding[1], 4 + padding[2])
    for dim, pad in enumerate(padding):
        if pad > 0:
            assert torch.allclose(
                padded_tensor.select(dim=dim, index=padded_tensor.size(dim) - 1),
                torch.tensor(fill_value),
            )


@pytest.mark.parametrize(
    ["padding"],
    [
        pytest.param([0, 0, 0], id="no_padding"),
        pytest.param([1, 1, 1], id="padding_1"),
    ],
)
def test_pad_unpad_rountrip(padding):
    tensor = torch.ones(2, 3, 4, device=get_device())
    padded_tensor = pad_tensor_at_end(tensor, padding)
    unpadded_tensor = unpad_tensor_at_end(padded_tensor, padding)
    assert unpadded_tensor.size() == tensor.size()
    assert torch.allclose(unpadded_tensor, tensor)


def run_gather_test(rank, worldsize):
    os.environ["MASTER_ADDR"] = "localhost"
    os.environ["MASTER_PORT"] = "12355"
    os.environ["WORLD_SIZE"] = str(worldsize)
    os.environ["RANK"] = str(rank)
    os.environ["LOCAL_RANK"] = str(rank)
    os.environ["NCCL_SHM_DISABLE"] = "1"

    dist = Distributed()
    assert dist.is_distributed()
    assert dist.rank == rank, f"Failed with Distrubuted rank {dist.rank} and arg {rank}"

    tensor = torch.ones(2, 5, device=get_device()) * dist.rank
    gathered = dist.gather(tensor)
    if dist.rank == 0:
        assert gathered is not None, "Gathered tensor are none instead of List"
        assert len(gathered) == dist.world_size
        for i in range(dist.world_size):
            assert torch.allclose(gathered[i].cpu(), torch.ones(2, 5) * i)


@pytest.mark.skipif(torch.cuda.device_count() < 2, reason="requires multi-GPU machine")
def test_distributed_gather():
    world_size = 2
    mp.spawn(run_gather_test, args=(world_size,), nprocs=world_size, join=True)


def test_non_distributed_gather():
    dist = Distributed()
    assert not dist.is_distributed()
    tensor = torch.ones(2, 5, device=get_device()) * 5
    gathered = dist.gather(tensor)
    assert gathered is not None, "Gathered tensor are none instead of List"
    assert len(gathered) == 1
    assert torch.allclose(gathered[0], tensor)



"""The registry also performs configuration set up so it needs to be tested."""

import copy
import datetime
from pathlib import Path
from unittest import mock

import numpy as np
import pytest
import torch

from fme.ace.stepper import SingleModuleStepperConfig, Stepper, parameter_init
from fme.core.coordinates import HybridSigmaPressureCoordinate, LatLonCoordinates
from fme.core.dataset_info import DatasetInfo
from fme.core.device import get_device
from fme.core.training_history import TrainingHistory
from fme.core.typing_ import TensorMapping
from fme.core.wildcard import wildcard_match

TIMESTEP = datetime.timedelta(hours=6)


def get_dataset_info(img_shape=(16, 32)) -> DatasetInfo:
    horizontal_coordinate = LatLonCoordinates(
        lat=torch.zeros(img_shape[0], device=get_device()),
        lon=torch.zeros(img_shape[1], device=get_device()),
    )
    return DatasetInfo(
        horizontal_coordinates=horizontal_coordinate,
        vertical_coordinate=HybridSigmaPressureCoordinate(
            ak=torch.arange(7), bk=torch.arange(7)
        ),
        timestep=TIMESTEP,
    )


def test_builder_with_weights_loads_same_state(tmpdir):
    sfno_config_data = {
        "type": "SphericalFourierNeuralOperatorNet",
        "config": {
            "num_layers": 2,
            "embed_dim": 3,
        },
    }
    stepper_config_data = {
        "builder": sfno_config_data,
        "in_names": ["x"],
        "out_names": ["x"],
        "normalization": {
            "means": {"x": np.random.randn(1).item()},
            "stds": {"x": np.random.randn(1).item()},
        },
    }
    stepper_config = SingleModuleStepperConfig.from_state(stepper_config_data)
    stepper = stepper_config.get_stepper(dataset_info=get_dataset_info())
    torch.save(
        {
            "stepper": stepper.get_state(),
        },
        str(tmpdir / "weights.ckpt"),
    )
    parameter_init_config = parameter_init.ParameterInitializationConfig(
        weights_path=str(tmpdir / "weights.ckpt"),
    )
    with_builder_stepper_config_data = {
        "builder": sfno_config_data,
        "in_names": ["x"],
        "out_names": ["x"],
        "normalization": {
            "means": {"x": np.random.randn(1).item()},
            "stds": {"x": np.random.randn(1).item()},
        },
        "parameter_init": parameter_init_config,
    }
    dataset_info = get_dataset_info()
    with_builder_stepper = SingleModuleStepperConfig.from_state(
        with_builder_stepper_config_data
    ).get_stepper(dataset_info=dataset_info)
    assert len(with_builder_stepper.modules) == 1
    assert_same_state(
        with_builder_stepper.modules[0].state_dict(),
        stepper.modules[0].state_dict(),
        allow_larger=False,
    )


def assert_same_state(
    state1: TensorMapping,
    state2: TensorMapping,
    allow_larger: bool,
    same_keys: bool = True,
):
    """
    Assert that two states are the same.

    If allow_larger is True then the states
    may have different dimension lengths, and if they do then the initial slice
    along that dimension will be compared.

    Args:
        state1: The first state.
        state2: The second state.
        allow_larger: Whether the states can have different dimension lengths.
        same_keys: Whether the states must have the same keys.
    """
    if same_keys:
        assert state1.keys() == state2.keys()
    same_state = []
    different_state = []
    for key in state1:
        if key in state2:
            if allow_larger:
                s = [
                    slice(0, min(s1, s2))
                    for s1, s2 in zip(state1[key].shape, state2[key].shape)
                ]
            else:
                s = [slice(0, None) for _ in state1[key].shape]
            if torch.allclose(state1[key][s], state2[key][s]):
                if not torch.all(state1[key] == 0):
                    same_state.append(key)
            else:
                different_state.append(key)
    assert len(same_state) > 0, (
        "No parameters were the same between the two states, "
        "or all parameters were uninitialized."
    )
    assert len(different_state) == 0, (
        f"Parameters {different_state} were different between "
        f"the two states, only {same_state} had the same non-zero state."
    )


@pytest.mark.parametrize(
    "loaded_shape,built_shape,extra_built_layer,expect_exception",
    [
        pytest.param((8, 16), (8, 16), False, False, id="same_shape"),
        pytest.param((8, 16), (8, 16), True, False, id="extra_layer"),
        pytest.param((8, 16), (16, 32), False, False, id="larger_model"),
        pytest.param((16, 32), (8, 16), False, True, id="smaller_model"),
    ],
)
def test_builder_with_weights_sfno_init(
    tmpdir, loaded_shape, built_shape, extra_built_layer: bool, expect_exception: bool
):
    """
    Integration test for the BuilderWithWeights stepper with a SFNO.
    """
    with_builder_stepper_config_data, _, stepper = get_config(
        loaded_shape, extra_built_layer, tmpdir
    )
    dataset_info = get_dataset_info(img_shape=built_shape)
    if expect_exception:
        with pytest.raises(ValueError):
            with_builder_stepper = SingleModuleStepperConfig.from_state(
                with_builder_stepper_config_data
            ).get_stepper(
                dataset_info=dataset_info,
            )
    else:
        with_builder_stepper = SingleModuleStepperConfig.from_state(
            with_builder_stepper_config_data
        ).get_stepper(
            dataset_info=dataset_info,
        )
        assert len(with_builder_stepper.modules) == 1
        if extra_built_layer:
            with pytest.raises(AssertionError):
                assert_same_state(
                    with_builder_stepper.modules[0].state_dict(),
                    stepper.modules[0].state_dict(),
                    allow_larger=True,
                    same_keys=True,  # This should fail if there are more parameters
                )
        assert_same_state(
            with_builder_stepper.modules[0].state_dict(),
            stepper.modules[0].state_dict(),
            allow_larger=True,
            same_keys=(not extra_built_layer),
        )


def get_config(
    loaded_shape: tuple[int, int],
    extra_built_layer: bool,
    tmpdir: Path,
):
    sfno_config_data = {
        "type": "SphericalFourierNeuralOperatorNet",
        "config": {
            "num_layers": 2,
            "embed_dim": 3,
            "scale_factor": 1,
        },
    }
    stepper_config_data = {
        "builder": sfno_config_data,
        "in_names": ["x"],
        "out_names": ["x"],
        "normalization": {
            "means": {"x": np.random.randn(1).item()},
            "stds": {"x": np.random.randn(1).item()},
        },
    }
    stepper_config = SingleModuleStepperConfig.from_state(stepper_config_data)
    dataset_info = get_dataset_info(img_shape=loaded_shape)
    stepper = stepper_config.get_stepper(dataset_info=dataset_info)
    built_sfno_config_data = copy.deepcopy(sfno_config_data)
    if extra_built_layer:
        built_sfno_config_data["config"]["num_layers"] += 1  # type: ignore
    torch.save(
        {
            "stepper": stepper.get_state(),
        },
        str(tmpdir / "weights.ckpt"),
    )
    parameter_init_config = parameter_init.ParameterInitializationConfig(
        weights_path=str(tmpdir / "weights.ckpt"),
    )
    with_builder_stepper_config_data = {
        "builder": built_sfno_config_data,
        "in_names": ["x"],
        "out_names": ["x"],
        "normalization": {
            "means": {"x": np.random.randn(1).item()},
            "stds": {"x": np.random.randn(1).item()},
        },
        "parameter_init": parameter_init_config,
    }
    return (
        with_builder_stepper_config_data,
        dataset_info,
        stepper,
    )


def test_with_weights_saved_stepper_does_not_need_untuned_weights(tmpdir):
    img_shape = (16, 32)
    with_builder_stepper_config_data, dataset_info, stepper = get_config(
        loaded_shape=img_shape, extra_built_layer=False, tmpdir=tmpdir
    )
    with_builder_stepper = SingleModuleStepperConfig.from_state(
        with_builder_stepper_config_data
    ).get_stepper(dataset_info=dataset_info)
    stepper_state = with_builder_stepper.get_state()
    # should be able to initialize stepper from its state without the untuned weights
    (tmpdir / "weights.ckpt").remove()
    stepper = Stepper.from_state(stepper_state)
    assert isinstance(stepper, Stepper)


class SimpleLinearModule(torch.nn.Module):
    def __init__(self, in_features, out_features):
        super().__init__()
        self.linear = torch.nn.Linear(in_features, out_features)
        self.custom_param = torch.nn.Parameter(torch.randn(out_features))

    def forward(self, x):
        return self.linear(x) + self.custom_param


class ComplexModule(torch.nn.Module):
    def __init__(self, in_features, out_features):
        super().__init__()
        self.linear1 = SimpleLinearModule(in_features, out_features)
        self.linear2 = SimpleLinearModule(out_features, in_features)
        self.custom_param = torch.nn.Parameter(torch.randn(in_features))

    def forward(self, x):
        return self.linear2(self.linear1(x)) + self.custom_param


@pytest.mark.parametrize(
    "apply_config",
    [
        pytest.param(True, id="frozen"),
        pytest.param(False, id="unfrozen"),
    ],
)
def test_frozen_parameter_config(apply_config: bool):
    module = ComplexModule(10, 20)
    config = parameter_init.FrozenParameterConfig(
        include=[
            "linear2.*",
            "custom_param",
            "linear1.custom_param",
            "linear1.linear.bias",
        ],
        exclude=["linear1.linear.weight"],
    )
    if apply_config:
        config.apply(module)
    # do some optimization and check only unfrozen parameters change
    original_state = copy.deepcopy(module.state_dict())
    optimizer = torch.optim.Adam(module.parameters())
    for _ in range(10):
        optimizer.zero_grad()
        loss = torch.sum(module(torch.randn(10, 10)))
        loss.backward()
        optimizer.step()
    for name, param in module.named_parameters():
        if name in config.exclude:
            assert not torch.allclose(param.data, original_state[name])
        else:
            if apply_config:
                assert torch.allclose(param.data, original_state[name])
            else:
                assert not torch.allclose(param.data, original_state[name])


@pytest.mark.parametrize(
    "include, exclude, expect_exception",
    [
        pytest.param(["*"], ["*"], True, id="both"),
        pytest.param(["*"], [], False, id="include"),
        pytest.param([], ["*"], False, id="exclude"),
        pytest.param(["linear1.*"], ["linear1.*"], True, id="both_same"),
        pytest.param(["linear1.*"], ["linear2.*"], False, id="both_different"),
        pytest.param(["linear1.*"], [], False, id="include"),
        pytest.param([], ["linear1.*"], False, id="exclude"),
        pytest.param(["linear1.*.weight"], ["linear1.*"], True, id="internal_wildcard"),
    ],
)
def test_frozen_parameter_config_no_overlaps(include, exclude, expect_exception):
    if expect_exception:
        with pytest.raises(ValueError):
            parameter_init.FrozenParameterConfig(include=include, exclude=exclude)
    else:
        parameter_init.FrozenParameterConfig(include=include, exclude=exclude)


def test_parameter_init_with_regularizer(tmpdir):
    """
    Test that optimizing on the parameter init regularizer will
    reduce the magnitude of randomly initialized weights, and reduce
    the distance between the weights and the original state for
    weights initialized from that state.
    """
    torch.manual_seed(0)
    device = get_device()
    saved_module = ComplexModule(10, 20).to(device)
    weights_path = str(tmpdir / "weights.ckpt")
    config = parameter_init.ParameterInitializationConfig(
        weights_path=weights_path,
        parameters=[
            parameter_init.ParameterClassification(
                exclude=["linear1.linear.weight"],
            ),
        ],
        alpha=1.0,
        beta=1.0,
    )
    parameter_initializer = config.build(
        load_weights_and_history=lambda _: (
            [saved_module.state_dict()],
            TrainingHistory(),
        )
    )
    # new_module = module
    module = ComplexModule(10, 20).to(device)
    parameter_initializer.apply_weights(
        [module],
    )
    regularizer = parameter_initializer.get_l2_sp_tuning_regularizer(
        modules=[module],
    )

    original_state = copy.deepcopy(module.state_dict())
    # overwrite new_module weights with random values
    for name, param in module.named_parameters():
        param.data[:] = torch.randn_like(param.data, device=device)
    pre_step_state = copy.deepcopy(module.state_dict())
    optimizer = torch.optim.SGD(module.parameters(), lr=0.01)

    optimizer.zero_grad()
    loss = regularizer()
    loss.backward()
    optimizer.step()

    for name, param in module.named_parameters():
        if any(
            wildcard_match(pattern, name) for pattern in config.parameters[0].exclude
        ):
            # L2 regularization against 0
            assert torch.all(param.data**2 < pre_step_state[name] ** 2)
        else:
            # L2 regularization against initial value
            assert torch.all(
                (param.data - original_state[name]) ** 2
                < (pre_step_state[name] - original_state[name]) ** 2
            )


def test_parameter_init_weights_loaded_once(tmpdir):
    """
    Test that the parameter initialization config only loads the weights once,
        even if multiple methods are called.
    """
    mock_load_weights_and_history = mock.Mock()
    mock_load_weights_and_history.return_value = (
        [torch.nn.Module().state_dict()],
        TrainingHistory(),
    )
    weights_path = str(tmpdir / "weights.ckpt")
    module = SimpleLinearModule(10, 20).to(get_device())
    config = parameter_init.ParameterInitializationConfig(
        weights_path=weights_path,
        parameters=[],
    )
    parameter_initializer = config.build(
        load_weights_and_history=mock_load_weights_and_history
    )
    # initially, the weights and history should not be loaded
    assert parameter_initializer._base_weights is None
    assert parameter_initializer._training_history is None
    mock_load_weights_and_history.assert_not_called
    # now call a method that uses the weights
    parameter_initializer.apply_weights(
        modules=[module],
    )
    # the weights should be loaded
    mock_load_weights_and_history.assert_called_once()
    # access the attributes that holds the weights and history
    assert parameter_initializer.base_weights is not None
    assert parameter_initializer.training_history is not None
    # the weights should not be loaded again
    mock_load_weights_and_history.assert_called_once()
    # call another method that uses the weights
    parameter_initializer.get_l2_sp_tuning_regularizer(
        modules=[module],
    )
    # the weights should not be loaded again
    mock_load_weights_and_history.assert_called_once()



import numpy as np


def u_v_to_x_y_z_wind(
    u: np.ndarray, v: np.ndarray, lat: np.ndarray, lon: np.ndarray
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Converts u and v wind components to x, y, z wind components.

    The x-axis is defined as the vector from the center of the earth to the
    intersection of the equator and the prime meridian. The y-axis is defined
    as the vector from the center of the earth to the intersection of the
    equator and the 90 degree east meridian. The z-axis is defined as the
    vector from the center of the earth to the north pole.

    Args:
        u: u wind component
        v: v wind component
        lat: latitude, in degrees
        lon: longitude, in degrees

    Returns:
        wx: x wind component
        wy: y wind component
        wz: z wind component
    """
    # for a graphical proof of the equations used here, see
    # https://github.com/ai2cm/full-model/pull/355#issuecomment-1729773301

    # Convert to radians
    lat = np.deg2rad(lat)
    lon = np.deg2rad(lon)

    # Horizontal winds
    #
    # Contribution from u
    #
    # The u component of the wind is aligned with the longitude lines,
    # so all of its contributions are to the x and y components of the wind.
    # At the prime meridian (which lies on the x axis), the u component points
    # parallel to the y-axis, so it contributes only to the y component of the
    # wind. As we move eastward, the u component points more and more in the
    # negative x direction, so it contributes more and more to the x component
    # of the wind. At the 90 degree east meridian, the u component points
    # parallel to the x-axis (in the negative direction).
    #
    # This influence on the x component is captured by multiplying the u component
    # by the negative sine of the longitude, which is zero at the prime meridian
    # and then becomes negative until reaching -1 at 90 degrees east.
    #
    # The influence on the y component is captured by multiplying the u component
    # by the cosine of the longitude, which is 1 at the prime meridian and then
    # decreases until reaching 0 at 90 degrees east.
    #
    # Contribution from v
    #
    # The v component of the wind is aligned with the latitude lines,
    # with no contribution to the horizontal at the equator and full contribution
    # at the poles. This is captured by multiplying the v component by the sine
    # of the latitude, which is 0 at the equator and 1 at the poles.
    #
    # The direction of the horizontal contribution is the vector pointing inwards
    # towards the axis of rotation of the Earth. At the prime meridian, this
    # is the negative x direction. At the 90 degree east meridian, this is the
    # negative y direction.
    #
    # This influence on the x component is captured by multiplying the v component
    # by the negative cosine of the longitude, which is 1 at the prime meridian
    # and then decreases until reaching 0 at 90 degrees east.
    #
    # The influence on the y component is captured by multiplying the v component
    # by the negative sine of the longitude, which is 0 at the prime meridian
    # and then decreases until reaching -1 at 90 degrees east.
    #
    # An exact derivation proving this effect is captured by sine and cosine
    # can be done graphically. Generally for these kinds of problems on a sphere
    # or circle it's always sine and cosine, and the question is which one and
    # whether it's positive or negative.

    # Wind in the x-direction:
    wx = -u * np.sin(lon) - v * np.sin(lat) * np.cos(lon)

    # Wind in the y-direction:
    wy = u * np.cos(lon) - v * np.sin(lat) * np.sin(lon)

    # Wind in the z-direction:

    # As the u-component is along latitude lines, and latitude lines are
    # perpendicular to Earth's axis of rotation, u does not appear in wz.
    #
    # The v-component of the wind is entirely aligned with Earth's axis of rotation
    # at the equator, and is entirely perpendicular at the poles, an effect that
    # is captured by multiplying the v component by the cosine of the latitude.
    wz = v * np.cos(lat)

    return wx, wy, wz


def normalize_vector(*vector_components: np.ndarray) -> np.ndarray:
    """
    Normalize a vector.

    The vector is assumed to be represented in an orthonormal basis, where
    each component is orthogonal to the others.

    Args:
        vector_components: components of the vector (e.g. x-, y-, and z-components)

    Returns:
        normalized vector, as a numpy array where each component has been
            concatenated along a new first dimension
    """
    scale = np.divide(
        1.0,
        np.sum(np.asarray([item**2.0 for item in vector_components]), axis=0) ** 0.5,
    )
    return np.asarray([item * scale for item in vector_components])


def lon_lat_to_xyz(lon, lat):
    """
    Convert (lon, lat) to (x, y, z).

    Args:
        lon: n-dimensional array of longitudes, in degrees
        lat: n-dimensional array of latitudes, in degrees

    Returns:
        x: n-dimensional array of x values
        y: n-dimensional array of y values
        z: n-dimensional array of z values
    """
    lat = np.deg2rad(lat)
    lon = np.deg2rad(lon)
    x = np.cos(lat) * np.cos(lon)
    y = np.cos(lat) * np.sin(lon)
    z = np.sin(lat)
    x, y, z = normalize_vector(x, y, z)
    return x, y, z


def xyz_to_lon_lat(x, y, z):
    """
    Convert (x, y, z) to (lon, lat).

    Args:
        x: n-dimensional array of x values
        y: n-dimensional array of y values
        z: n-dimensional array of z values

    Returns:
        lon: n-dimensional array of longitudes, in degrees
        lat: n-dimensional array of latitudes, in degrees
    """
    x, y, z = normalize_vector(x, y, z)
    # double transpose to index last dimension, regardless of number of dimensions
    lon = np.zeros_like(x)
    nonzero_lon = np.abs(x) + np.abs(y) >= 1.0e-10
    lon[nonzero_lon] = np.arctan2(y[nonzero_lon], x[nonzero_lon])
    negative_lon = lon < 0.0
    while np.any(negative_lon):
        lon[negative_lon] += 2 * np.pi
        negative_lon = lon < 0.0
    lat = np.arcsin(z)
    lat = np.rad2deg(lat)
    lon = np.rad2deg(lon)
    return lon, lat



# Compute the 1940-2021 ENSO/Nino 3.4 index from the CMIP6 AMIP SST forcing data
# Based on notebook here: github.com/ai2cm/explore/blob/master/brianh/2024-05-06-AMIP-interannual-variability/2024-07-03-compute-CMIP6-AMIP-ENSO-index.ipynb # noqa

import dataclasses
from datetime import timedelta

import numpy as np
import xarray as xr

# SST_SOURCE = "gs://vcm-ml-intermediate/2023-10-06-AMIP-forcing-data/sst.nc"
# !gsutil cp {SST_SOURCE} .
SST_DATASET = "./sst.nc"
OCEAN_MASK_SOURCE = "gs://vcm-ml-intermediate/2023-10-27-vertically-resolved-1deg-fme-amip-ensemble-dataset/ic_0001.zarr"  # noqa
NINO_REGION_BOUNDS = dict(lat=(-5, 5), lon=(190, 240))
TROPICAL_REGION_BOUNDS = dict(lat=(-5, 5), lon=(0, 360))
OUTPUT_FILE = "./index.py"


@dataclasses.dataclass
class RegionBounds:
    lat: tuple[int, int]
    lon: tuple[int, int]

    def get_regional_average(self, ds, weights):
        masked = (
            ds.where(ds.lat >= self.lat[0])
            .where(ds.lat <= self.lat[1])
            .where(ds.lon >= self.lon[0])
            .where(ds.lon <= self.lon[1])
        )
        return masked.weighted(weights).mean(dim=["lat", "lon"])


def get_ocean_mask(source, template: xr.DataArray):
    """
    Get an ocean mask from the FV3GFS dataset. Regrid to the 1deg lat/lon grid
    of the SST dataset.
    """
    ds = xr.open_zarr(source)
    ocean_fraction = ds.ocean_fraction.isel(time=-1).rename(
        {"grid_xt": "lon", "grid_yt": "lat"}
    )
    ocean_fraction_1deg = ocean_fraction.interp_like(template).load()
    ocean_mask = xr.where(ocean_fraction_1deg > 0.5, True, False)
    return ocean_mask


def get_time_average(da):
    # this version of xarray's resample method doesn't allow
    # data shifting to create a centered 3-month average, so do it manually
    da = da.assign_coords({"time": da.time + timedelta(days=45)})
    # the label is at the start of the 3-month season
    da_out = da.resample(time="QS").mean()
    return da_out


def get_anomalies(
    ds,
    time_groupby_key,
    time_dim="time",
    start_time="1940-01-01",
    end_time="2021-01-01",
):
    ds = ds.sel(time=slice(start_time, end_time))
    full_groupby_key = f"{time_dim}.{time_groupby_key}"
    ds_monthly_climo = ds.groupby(full_groupby_key).mean(time_dim)
    ds_monthly_anomalies = ds.groupby(full_groupby_key) - ds_monthly_climo
    return ds_monthly_anomalies


def get_time_trendline(da):
    coeff = np.polyfit(np.arange(da.sizes["time"]), da.values, deg=1)
    da_trend = xr.DataArray(
        coeff[0] * np.arange(da.sizes["time"]) + coeff[1],
        dims="time",
        coords={"time": da.time},
    )
    return da_trend


def main():
    surface_temperature = xr.open_dataset(SST_DATASET)["sea_surface_temperature"]
    ocean_mask = get_ocean_mask(OCEAN_MASK_SOURCE, template=surface_temperature)
    area_weights = np.cos(np.deg2rad(surface_temperature.lat))
    nino34_region = RegionBounds(**NINO_REGION_BOUNDS)
    tropical_region = RegionBounds(**TROPICAL_REGION_BOUNDS)

    nino34_temperature = nino34_region.get_regional_average(
        surface_temperature, area_weights
    )
    tropical_sst = tropical_region.get_regional_average(
        surface_temperature, weights=(area_weights * ocean_mask)
    )
    nino34_temperature_anom = nino34_temperature - tropical_sst
    nino34_temperature_anom_index = get_anomalies(
        nino34_temperature_anom, time_groupby_key="month"
    )
    nino34_anom_index = get_time_average(nino34_temperature_anom_index)

    with open(OUTPUT_FILE, "w") as f:
        print(
            (
                "# Nino3.4 index anomaly from tropical SST average, "
                "3-monthly centered running mean [K]"
            ),
            file=f,
        )
        print(
            "# computed from the CMIP6 AMIP SST forcing data using the script in",
            file=f,
        )
        print("# `scripts/monthly_data/compute_enso_index.py`", file=f)
        print("NINO34_INDEX = [", file=f)
        for point in nino34_anom_index:
            print(
                (
                    f"    (({point.time.item().year}, {point.time.item().month}, "
                    f"{point.time.item().day}), {point.item():0.3f}),"
                ),
                file=f,
            )
        print("]", file=f)


if __name__ == "__main__":
    main()



import re

import pytest
import torch

import fme
from fme.core.masking import StaticMasking, StaticMaskingConfig

DEVICE = fme.get_device()


class _Mask:
    LEVEL_PATTERN = re.compile(r"_(\d+)$")

    def __init__(
        self,
        mask_2d: torch.Tensor | None = None,
        mask_3d: torch.Tensor | None = None,
    ):
        self.mask_2d = mask_2d
        self.mask_3d = mask_3d

    def get_mask_tensor_for(self, name: str) -> torch.Tensor | None:
        if name == "mask_ignored":
            return None
        match = self.LEVEL_PATTERN.search(name)
        if match:
            # 3D variable
            if self.mask_3d is None:
                return None
            level = int(match.group(1))
            return self.mask_3d.select(dim=-1, index=level)
        else:
            # 2D variable
            return self.mask_2d

    def build_output_masker(self):
        raise NotImplementedError

    def to(self, device: str) -> "_Mask":
        return self


def test_masking_config():
    config = StaticMaskingConfig(
        mask_value=1,
        fill_value=0.0,
    )
    _ = config.build(_Mask(mask_2d=torch.ones(1, 1)))
    _ = config.build(_Mask(mask_3d=torch.ones(1, 1, 1)))
    _ = config.build(_Mask(mask_2d=torch.ones(1, 1), mask_3d=torch.ones(1, 1, 1)))

    with pytest.raises(ValueError, match="mask_value must be either 0 or 1"):
        _ = StaticMaskingConfig(
            mask_value=3,
            fill_value=0.0,
        )

    config = StaticMaskingConfig(
        mask_value=1,
        fill_value="mean",
    )
    with pytest.raises(ValueError, match="fill_values mapping required"):
        _ = config.build(_Mask(mask_2d=torch.ones(1, 1)))


_SIZE = (4, 4)

_MASK_2D = torch.ones(_SIZE, device=DEVICE)
_MASK_3D = torch.ones(_SIZE + (2,), device=DEVICE)
_MASK_2D[1, 1] = 0
_MASK_3D[0, :, 0] = 0
_MASK_3D[1, :, 1] = 0


_DATA = {
    "PRESsfc": 10.0 + torch.rand(size=_SIZE, device=DEVICE),
    "specific_total_water_0": torch.rand(size=_SIZE, device=DEVICE),
    "specific_total_water_1": torch.rand(size=_SIZE, device=DEVICE),
}


@pytest.mark.parametrize(
    "exclude",
    [
        None,
        [],
        ["specific_total_water_1"],
    ],
)
def test_masking(exclude):
    config = StaticMaskingConfig(
        mask_value=0,
        fill_value=0.0,
        exclude_names_and_prefixes=exclude,
    )
    mask = config.build(_Mask(_MASK_2D, _MASK_3D))
    output = mask(_DATA)
    assert output["PRESsfc"][1, 1] == 0.0
    assert output["PRESsfc"][0, 1] != 0.0
    assert torch.all(output["specific_total_water_0"][0, :] == 0.0)
    if exclude and "specific_total_water_1" in exclude:
        assert torch.all(output["specific_total_water_1"][1, :] != 0.0)
    else:
        assert torch.all(output["specific_total_water_1"][1, :] == 0.0)
    assert torch.all(output["specific_total_water_1"][0, :] != 0.0)


@pytest.mark.parametrize(
    "exclude",
    [
        "PRESsfc",
        "specific_total_water_",
        "specific_total_water",
        "specific_total_water_0",
    ],
)
def test_masking_exclusion(exclude):
    config = StaticMaskingConfig(
        mask_value=0,
        fill_value=float("nan"),
        exclude_names_and_prefixes=[exclude],
    )
    mask = config.build(mask=_Mask(_MASK_2D, _MASK_3D))
    output = mask(_DATA)
    if exclude == "PRESsfc":
        assert torch.all(~output["PRESsfc"].isnan())
        assert torch.all(output["specific_total_water_0"][0, :].isnan())
        assert torch.all(output["specific_total_water_1"][1, :].isnan())
    else:
        assert output["PRESsfc"][1, 1].isnan()
    if exclude == "specific_total_water_0":
        assert torch.all(~output["specific_total_water_0"].isnan())
        assert torch.all(output["specific_total_water_1"][1, :].isnan())
    elif exclude != "PRESsfc":
        assert torch.all(~output["specific_total_water_0"].isnan())
        assert torch.all(~output["specific_total_water_1"].isnan())


def test_masking_with_means():
    config = StaticMaskingConfig(
        mask_value=0,
        fill_value="mean",
    )
    mask = config.build(
        mask=_Mask(_MASK_2D, _MASK_3D),
        means={
            "PRESsfc": torch.tensor(1.0, device=DEVICE),
            "specific_total_water_0": torch.tensor(2.0, device=DEVICE),
            "specific_total_water_1": torch.tensor(3.0, device=DEVICE),
        },
    )
    output = mask(_DATA)
    assert output["PRESsfc"][1, 1] == 1.0
    assert output["PRESsfc"][0, 1] != 1.0
    assert torch.all(output["specific_total_water_0"][0, :] == 2.0)
    assert torch.all(output["specific_total_water_1"][1, :] == 3.0)
    assert torch.all(output["specific_total_water_1"][0, :] != 3.0)


def test_masking_no_3d_masking():
    config = StaticMaskingConfig(
        mask_value=0,
        fill_value=0.0,
        exclude_names_and_prefixes=["specific_total_water"],
    )
    mask = config.build(_Mask(mask_2d=_MASK_2D))
    output = mask(_DATA)
    assert output["PRESsfc"][1, 1] == 0.0
    assert output["PRESsfc"][0, 1] != 0.0
    assert torch.all(output["specific_total_water_0"][0, :] != 0.0)
    assert torch.all(output["specific_total_water_1"][1, :] != 0.0)
    assert torch.all(output["specific_total_water_1"][0, :] != 0.0)


def test_masking_no_surface_masking():
    config = StaticMaskingConfig(
        mask_value=0,
        fill_value=0.0,
        exclude_names_and_prefixes=["PRESsfc"],
    )
    mask = config.build(_Mask(mask_3d=_MASK_3D))
    output = mask(_DATA)
    assert output["PRESsfc"][1, 1] != 0.0
    assert output["PRESsfc"][0, 1] != 0.0
    assert torch.all(output["specific_total_water_0"][0, :] == 0.0)
    assert torch.all(output["specific_total_water_1"][1, :] == 0.0)
    assert torch.all(output["specific_total_water_1"][0, :] != 0.0)


def test_masking_missing_2d_mask():
    config = StaticMaskingConfig(
        mask_value=0,
        fill_value=0.0,
        exclude_names_and_prefixes=["specific_total_water"],
    )
    mask = config.build(_Mask(mask_3d=_MASK_3D))
    masked = mask(_DATA)
    torch.testing.assert_close(masked["PRESsfc"], _DATA["PRESsfc"])


def test_masking_missing_3d_mask():
    config = StaticMaskingConfig(
        mask_value=0,
        fill_value=0.0,
    )
    mask = config.build(_Mask(mask_2d=_MASK_2D))
    masked = mask(_DATA)
    assert masked["PRESsfc"][1, 1] == 0.0
    assert masked["PRESsfc"][0, 1] != 0.0
    for name in ["specific_total_water_0", "specific_total_water_1"]:
        torch.testing.assert_close(masked[name], _DATA[name])


def test_static_masking_error_on_missing_mean():
    mask = StaticMasking(
        mask_value=0,
        fill_value={
            "specific_total_water_0": torch.tensor(2.0, device=DEVICE),
            "specific_total_water_1": torch.tensor(3.0, device=DEVICE),
        },
        mask=_Mask(_MASK_2D, _MASK_3D),
    )
    with pytest.raises(KeyError, match="missing key 'PRESsfc'"):
        _ = mask(_DATA)


def test_static_masking_mask_ignored_name():
    mask = StaticMasking(
        mask_value=0,
        fill_value=float("nan"),
        mask=_Mask(_MASK_2D, _MASK_3D),
    )
    data = {
        "masked": torch.rand(size=_SIZE, device=DEVICE),
        "mask_ignored": torch.rand(size=_SIZE, device=DEVICE),
    }
    masked = mask(data)
    assert masked["masked"][1, 1].isnan()
    # no change to "mask_ignored" because _Mask gives it special treatment
    torch.testing.assert_close(masked["mask_ignored"], data["mask_ignored"])



import abc
from typing import Any, Generic, TypeVar

PS = TypeVar("PS", contravariant=True)  # prognostic state
T = TypeVar("T", contravariant=True)


class AggregatorABC(abc.ABC, Generic[T]):
    @abc.abstractmethod
    def record_batch(self, batch: T) -> None:
        pass

    @abc.abstractmethod
    def get_logs(self, label: str) -> dict[str, float]:
        pass

    @abc.abstractmethod
    def flush_diagnostics(self, subdir: str | None) -> None:
        pass


InferenceLog = dict[str, Any]
InferenceLogs = list[InferenceLog]


class InferenceAggregatorABC(abc.ABC, Generic[PS, T]):
    @abc.abstractmethod
    def record_batch(
        self,
        data: T,
    ) -> InferenceLogs:
        """
        Record a batch of data.

        Args:
            data: Batch of data.

        Returns:
            Logs for the batch.
        """
        pass

    @abc.abstractmethod
    def record_initial_condition(
        self,
        initial_condition: PS,
    ) -> InferenceLogs:
        """
        Record the initial condition.

        May only be recorded once, before any calls to record_batch.
        """
        pass

    @abc.abstractmethod
    def get_summary_logs(self) -> InferenceLog:
        pass

    @abc.abstractmethod
    def flush_diagnostics(self, subdir: str | None) -> None:
        pass



import xarray as xr
from combine_stats import get_combined_stats


def test_get_combined_stats():
    import numpy as np

    arr1 = xr.DataArray(np.random.rand(10, 10), dims=["x", "y"])
    arr2 = xr.DataArray(np.random.rand(5, 10) * 2 + 5, dims=["x", "y"])
    full_field_datasets = [
        xr.Dataset(
            {"arr": arr1.std()},
            attrs={"input_samples": 100},
        ),
        xr.Dataset(
            {"arr": arr2.std()},
            attrs={"input_samples": 50},
        ),
    ]
    centering_datasets = [
        xr.Dataset(
            {"arr": arr1.mean()},
            attrs={"input_samples": 100},
        ),
        xr.Dataset(
            {"arr": arr2.mean()},
            attrs={"input_samples": 50},
        ),
    ]
    samples = xr.DataArray([100, 50], dims=["run"])
    average = get_combined_stats(full_field_datasets, centering_datasets, samples)
    combined_arr = np.concatenate([arr1.values.flatten(), arr2.values.flatten()])
    assert np.allclose(average["arr"], np.std(combined_arr))



import pytest
import torch

from fme.core.tensors import (
    assert_dict_allclose,
    fold_sized_ensemble_dim,
    repeat_interleave_batch_dim,
    unfold_ensemble_dim,
)
from fme.core.typing_ import EnsembleTensorDict


@pytest.mark.parametrize(
    "a, b",
    [
        (
            {"a": torch.tensor(1)},
            {"a": torch.tensor(1)},
        ),
        (
            {"a": torch.tensor(1), "b": {"c": torch.tensor(2)}},
            {"a": torch.tensor(1), "b": {"c": torch.tensor(2)}},
        ),
    ],
)
def test_assert_dict_allclose_passes_for_equal_dicts(a, b):
    assert_dict_allclose(a, b)


@pytest.mark.parametrize(
    "a, b",
    [
        (
            {"a": torch.tensor(1), "b": {"c": torch.tensor(2)}},
            {"a": torch.tensor(1), "b": {"c": torch.tensor(3)}},
        ),
        (
            {"a": torch.tensor(1)},
            {"b": torch.tensor(1)},
        ),
        (
            {"a": torch.tensor(1)},
            {"a": torch.tensor(2)},
        ),
    ],
)
def test_assert_allclose_raises_for_unequal_dicts(a, b):
    with pytest.raises(AssertionError):
        assert_dict_allclose(a, b)


@pytest.mark.parametrize("repeats", [1, 3])
def test_ensemble_dim_operations_give_correct_data_order(repeats: int):
    n_samples = 3
    data = {
        "a": torch.arange(n_samples)
        .reshape(n_samples, 1, 1)
        .broadcast_to(n_samples, 5, 5),
        "b": torch.arange(n_samples)
        .reshape(n_samples, 1, 1)
        .broadcast_to(n_samples, 5, 5),
    }
    intermediate = repeat_interleave_batch_dim(data, repeats)
    for k in data:
        assert intermediate[k].shape == (n_samples * repeats, 5, 5)
    result = unfold_ensemble_dim(intermediate, repeats)
    for k in data:
        assert result[k].shape == (n_samples, repeats, 5, 5)
        assert torch.allclose(result[k], data[k][:, None, :, :])


def test_fold_sized_ensemble_dim_broadcasts_singleton():
    a = torch.randn(1, 1, 1)
    d = EnsembleTensorDict({"a": a})
    folded = fold_sized_ensemble_dim(d, 2)
    assert set(folded.keys()) == {"a"}
    assert folded["a"].shape == (2, 1)
    assert (folded["a"] == a).all()  # allclose doesn't broadcast


def test_fold_sized_ensemble_dim_reduces_dimensionality():
    a = torch.randn(2, 3, 4, 8)
    d = EnsembleTensorDict({"a": a})
    folded = fold_sized_ensemble_dim(d, 3)
    assert set(folded.keys()) == {"a"}
    assert folded["a"].shape == (6, 4, 8)
    torch.testing.assert_close(folded["a"], a.reshape(6, 4, 8))


def test_fold_sized_ensemble_dim_raises_on_invalid_ensemble_counts():
    a = torch.randn(2, 3, 4, 8)
    b = torch.randn(2, 4, 4, 8)
    d = EnsembleTensorDict({"a": a, "b": b})
    with pytest.raises(ValueError) as excinfo:
        fold_sized_ensemble_dim(d, 3)
    assert "some values in d have invalid ensemble member counts" in str(excinfo.value)
    assert "a: 3" in str(excinfo.value)
    assert "b: 4" in str(excinfo.value)


def test_unfold_ensemble_dim_increases_dimensionality():
    a = torch.randn(6, 4, 8)
    d = EnsembleTensorDict({"a": a})
    unfolded = unfold_ensemble_dim(d, 3)
    assert set(unfolded.keys()) == {"a"}
    assert unfolded["a"].shape == (2, 3, 4, 8)
    torch.testing.assert_close(unfolded["a"], a.reshape(2, 3, 4, 8))



import dataclasses
from typing import Protocol

import torch
import xarray as xr

TIME_DIM_NAME = "time"
TIME_DIM = 1  # sample, time, lat, lon


class _PairedDataWriter(Protocol):
    def append_batch(
        self,
        target: dict[str, torch.Tensor],
        prediction: dict[str, torch.Tensor],
        start_timestep: int,
        batch_time: xr.DataArray,
    ):
        pass

    def flush(self):
        pass


class _DataWriter(Protocol):
    def append_batch(
        self,
        data: dict[str, torch.Tensor],
        start_timestep: int,
        batch_time: xr.DataArray,
    ):
        pass

    def flush(self):
        pass


@dataclasses.dataclass
class TimeCoarsenConfig:
    """
    Config for inference data time coarsening.

    Args:
        coarsen_factor: Factor by which to coarsen in time, an integer 1 or greater. The
            resulting time labels will be coarsened to the mean of the original labels.
    """

    def __post_init__(self):
        if self.coarsen_factor < 1:
            raise ValueError(
                f"coarsen_factor must be 1 or greater, got {self.coarsen_factor}"
            )

    coarsen_factor: int

    def build_paired(self, data_writer: _PairedDataWriter) -> "PairedTimeCoarsen":
        return PairedTimeCoarsen(
            data_writer=data_writer,
            coarsen_factor=self.coarsen_factor,
        )

    def build(self, data_writer: _DataWriter) -> "TimeCoarsen":
        return TimeCoarsen(
            data_writer=data_writer,
            coarsen_factor=self.coarsen_factor,
        )

    def n_coarsened_timesteps(self, n_timesteps: int) -> int:
        """Assumes initial condition is NOT in n_timesteps."""
        return (n_timesteps) // self.coarsen_factor


class PairedTimeCoarsen:
    """Wraps a data writer and coarsens its arguments in time before passing them on."""

    def __init__(
        self,
        data_writer: _PairedDataWriter,
        coarsen_factor: int,
    ):
        self._data_writer: _PairedDataWriter = data_writer
        self._coarsen_factor: int = coarsen_factor

    def append_batch(
        self,
        target: dict[str, torch.Tensor],
        prediction: dict[str, torch.Tensor],
        start_timestep: int,
        batch_time: xr.DataArray,
    ):
        (target_coarsened, start_timestep, batch_times_coarsened) = coarsen_batch(
            target, start_timestep, batch_time, self._coarsen_factor
        )
        (prediction_coarsened, _, _) = coarsen_batch(
            prediction, start_timestep, batch_time, self._coarsen_factor
        )
        self._data_writer.append_batch(
            target_coarsened,
            prediction_coarsened,
            start_timestep,
            batch_times_coarsened,
        )

    def flush(self):
        self._data_writer.flush()


class TimeCoarsen:
    """Wraps a data writer and coarsens its arguments in time before passing them on."""

    def __init__(
        self,
        data_writer: _DataWriter,
        coarsen_factor: int,
    ):
        self._data_writer: _DataWriter = data_writer
        self._coarsen_factor: int = coarsen_factor

    def append_batch(
        self,
        data: dict[str, torch.Tensor],
        start_timestep: int,
        batch_time: xr.DataArray,
    ):
        (data_coarsened, start_timestep, batch_times_coarsened) = coarsen_batch(
            data, start_timestep, batch_time, self._coarsen_factor
        )
        self._data_writer.append_batch(
            data_coarsened,
            start_timestep,
            batch_times_coarsened,
        )

    def flush(self):
        self._data_writer.flush()


def coarsen_batch(
    data: dict[str, torch.Tensor],
    start_timestep: int,
    batch_time: xr.DataArray,
    coarsen_factor: int,
) -> tuple[dict[str, torch.Tensor], int, xr.DataArray]:
    data_coarsened = _coarsen_tensor_dict(data, coarsen_factor)
    start_timestep = start_timestep // coarsen_factor
    batch_time_coarsened = batch_time.coarsen({TIME_DIM_NAME: coarsen_factor}).mean()
    return data_coarsened, start_timestep, batch_time_coarsened


def _coarsen_tensor_dict(
    tensor_dict: dict[str, torch.Tensor], coarsen_factor: int
) -> dict[str, torch.Tensor]:
    """Coarsen each tensor along a given axis by a given factor."""
    coarsened_tensor_dict = {}
    for name, tensor in tensor_dict.items():
        coarsened_tensor_dict[name] = tensor.unfold(
            dimension=TIME_DIM, size=coarsen_factor, step=coarsen_factor
        ).mean(dim=-1)
    return coarsened_tensor_dict



import pytest

from fme.core.dicts import add_names, to_flat_dict, to_nested_dict


def get_cfg_and_args_dicts():
    config_d = {
        "top": 1,
        "seq": [dict(a=1), dict(a=2)],
        "nested": {"k1": 2, "k2": 3, "double_nest": {"k1": 4, "k2": 5}},
    }

    flat_d = {
        "top": 1,
        "seq": [dict(a=1), dict(a=2)],
        "nested.k1": 2,
        "nested.k2": 3,
        "nested.double_nest.k1": 4,
        "nested.double_nest.k2": 5,
    }

    return config_d, flat_d


def test_to_flat_dict():
    config_d, expected = get_cfg_and_args_dicts()
    result = to_flat_dict(config_d)
    assert result == expected


def test_to_nested_dict():
    expected, args_d = get_cfg_and_args_dicts()
    result = to_nested_dict(args_d)
    assert result == expected


def test_flat_dict_round_trip():
    config_d, _ = get_cfg_and_args_dicts()

    args_d = to_flat_dict(config_d)
    result = to_nested_dict(args_d)

    assert result == config_d


def test_add_names():
    input = {"a": 1, "b": 2}
    output = {"a": 3, "b": 4}
    names = ["a"]
    new_output = add_names(input, output, names)
    assert new_output["a"] == 4
    assert new_output["b"] == 4


def test_add_names_passes_with_extra_name():
    input = {"a": 1, "b": 2}
    output = {"a": 3, "b": 4}
    names = ["a", "c"]
    new_output = add_names(input, output, names)
    assert new_output["a"] == 4
    assert new_output["b"] == 4


def test_add_names_raises_key_error_on_missing_input():
    input = {"a": 1}
    output = {"a": 3, "b": 4}
    names = ["b"]
    with pytest.raises(KeyError):
        add_names(input, output, names)



import logging
import time

import numpy as np
import pytest

from fme.core.timing import CumulativeTimer, GlobalTimer


def test_CumulativeTimer():
    category = "foo"
    cumulative_timer = CumulativeTimer(category)

    cumulative_timer.start()
    time.sleep(0.01)
    cumulative_timer.stop()

    time.sleep(0.01)

    cumulative_timer.start()
    time.sleep(0.01)
    cumulative_timer.stop()

    assert cumulative_timer.duration == pytest.approx(0.02, abs=0.005)


def test_CumulativeTimer_start_error():
    category = "foo"
    cumulative_timer = CumulativeTimer(category)
    cumulative_timer.start()
    with pytest.raises(RuntimeError, match=f"timer {category!r} is already running"):
        cumulative_timer.start()


def test_CumulativeTimer_stop_error():
    category = "foo"
    cumulative_timer = CumulativeTimer(category)
    with pytest.raises(
        RuntimeError, match=f"must call start for timer {category!r} before stop"
    ):
        cumulative_timer.stop()


def test_CumulativeTimer_duration_error():
    category = "foo"
    cumulative_timer = CumulativeTimer(category)
    cumulative_timer.start()
    with pytest.raises(RuntimeError, match=f"timer {category!r} is still running"):
        cumulative_timer.duration


def exercise_active_timer():
    timer = GlobalTimer.get_instance()

    timer.start("foo")
    time.sleep(0.01)
    timer.stop()

    timer = GlobalTimer.get_instance()

    timer.start("bar")
    time.sleep(0.02)
    timer.stop()

    assert timer.get_duration("foo") == pytest.approx(0.01, abs=0.005)
    assert timer.get_duration("bar") == pytest.approx(0.02, abs=0.005)

    with pytest.raises(KeyError):
        timer.get_duration("baz")

    durations = timer.get_durations()
    assert durations["foo"] == pytest.approx(0.01, abs=0.005)
    assert durations["bar"] == pytest.approx(0.02, abs=0.005)


def test_GlobalTimer():
    with GlobalTimer():
        exercise_active_timer()

    # Check that timer is reset within new context.
    with GlobalTimer():
        timer = GlobalTimer.get_instance()
        assert timer.get_durations() == {}


def test_GlobalTimer_resets_after_exception():
    with pytest.raises(ValueError):
        with GlobalTimer():
            timer = GlobalTimer.get_instance()
            timer.start("foo")
            raise ValueError()

    # Check that the context manager clears the state of the timer after an
    # exception. If it were not clear, starting the timer for "foo" would raise
    # an error.
    with GlobalTimer():
        timer = GlobalTimer.get_instance()
        timer.start("foo")
        timer.stop()


def test_GlobalTimer_multiple_context_error():
    with pytest.raises(RuntimeError, match="GlobalTimer is currently in use"):
        with GlobalTimer(), GlobalTimer():
            pass


def test_inactive_GlobalTimer_warning():
    with pytest.warns(UserWarning, match=r"inactive"):
        GlobalTimer.get_instance()


@pytest.mark.filterwarnings("ignore:The GlobalTimer")
def test_inactive_GlobalTimer_start():
    timer = GlobalTimer.get_instance()
    timer.start("foo")


@pytest.mark.filterwarnings("ignore:The GlobalTimer")
def test_inactive_GlobalTimer_stop():
    timer = GlobalTimer.get_instance()
    timer.stop()


@pytest.mark.filterwarnings("ignore:The GlobalTimer")
def test_inactive_GlobalTimer_get_duration():
    timer = GlobalTimer.get_instance()
    result = timer.get_duration("foo")
    assert np.isnan(result)


@pytest.mark.filterwarnings("ignore:The GlobalTimer")
def test_inactive_GlobalTimer_get_durations():
    timer = GlobalTimer.get_instance()
    result = timer.get_durations()
    assert result == {}


@pytest.mark.filterwarnings("ignore:The GlobalTimer")
def test_inactive_GlobalTimer_log_durations(caplog):
    timer = GlobalTimer.get_instance()
    with caplog.at_level(logging.INFO):
        timer.log_durations()
    assert len(caplog.records) == 0


@pytest.mark.filterwarnings("ignore:The GlobalTimer")
def test_GlobalTimer_inactive_then_active():
    # Make sure we can instantiate an inactive timer, but then still create and
    # use an active timer later.
    GlobalTimer.get_instance()

    with GlobalTimer():
        exercise_active_timer()


def test_GlobalTimer_context():
    with GlobalTimer():
        timer = GlobalTimer.get_instance()
        with timer.context("foo"):
            time.sleep(0.01)
    assert timer.get_duration("foo") > 0.01


def test_GlobalTimer_context_with_exception():
    with pytest.raises(ValueError):
        with GlobalTimer():
            timer = GlobalTimer.get_instance()
            with timer.context("foo"):
                time.sleep(0.01)
                raise ValueError()
    assert timer.get_duration("foo") > 0.01


def test_GlobalTimer_single_inner_timer():
    with pytest.raises(
        RuntimeError, match="GlobalTimer already has an active inner timer"
    ):
        with GlobalTimer():
            timer = GlobalTimer.get_instance()
            with timer.context("foo"):
                with timer.context("bar"):
                    pass


def test_GlobalTimer_nested_outer_context():
    with GlobalTimer():
        timer = GlobalTimer.get_instance()
        with timer.outer_context("foo"):
            with timer.context("bar"):
                pass


def test_GlobalTimer_double_nested_outer_context():
    with GlobalTimer():
        timer = GlobalTimer.get_instance()
        with timer.outer_context("foo"):
            with timer.outer_context("bar"):
                pass
